import datetime
from typing import Any, Optional

from common.database import DBConfig, Db
from common.database.exceptions import UniqueConstraintError
from common.logging import get_logger, span
from common.models.common import (
    UserAccessToken,
    Platform,
    UserAccessTokenCreate,
    UserAccessTokenUpdate,
)
from common.models.user import (
    SocialMediaAccount,
    SocialMediaAccountCreate,
    User,
    UserCreate,
    UserUpdate,
    TwitchAccount,
    TwitchAccountCreate,
    TwitchAccountUpdate,
    KickAccount,
    KickAccountCreate,
    KickAccountUpdate,
)
from common.utils.crypto import Crypto, CryptoConfig

USER_COLUMNS_LEN = 13
USER_OAUTH_COLUMNS_LEN = 11
KICK_ACCOUNT_COLUMNS_LEN = 8
TWITCH_ACCOUNT_COLUMNS_LEN = 8
KICK_ACCOUNT_OAUTH_COLUMNS_LEN = 10
TWITCH_ACCOUNT_OAUTH_COLUMNS_LEN = 10

SELECT_ONLY_USER_QUERY = """
                         SELECT u.id,
                                u.first_promoter_id,
                                u.first_name,
                                u.last_name,
                                u.email,
                                u.phone,
                                u.tier,
                                u.product,
                                u.created_at,
                                u.updated_at,
                                u.deleted_at,
                                u.last_login,
                                u.customer_journey_completed_at
                         FROM users u \
                         """

SELECT_USER_QUERY = """
                    SELECT u.id,
                           u.first_promoter_id,
                           u.first_name,
                           u.last_name,
                           u.email,
                           u.phone,
                           u.tier,
                           u.product,
                           u.created_at,
                           u.updated_at,
                           u.deleted_at,
                           u.last_login,
                           u.customer_journey_completed_at,

                           -- Twitch account
                           ta.id                   AS twitch_account_id,
                           ta.twitch_id,
                           ta.twitch_username,
                           ta.profile_link         AS twitch_profile_link,
                           ta.profile_picture      AS twitch_profile_picture,
                           ta.created_at           AS twitch_created_at,
                           ta.updated_at           AS twitch_updated_at,
                           ta.deleted_at           AS twitch_deleted_at,

                           -- Kick account
                           ka.id                   AS kick_account_id,
                           ka.kick_id,
                           ka.kick_username,
                           ka.profile_link         AS kick_profile_link,
                           ka.profile_picture      AS kick_profile_picture,
                           ka.created_at           AS kick_created_at,
                           ka.updated_at           AS kick_updated_at,
                           ka.deleted_at           AS kick_deleted_at,

                           -- Twitch OAuth token
                           ut_twitch.id            AS twitch_oauth_id,
                           ut_twitch.platform      AS twitch_platform,
                           ut_twitch.access_token  AS twitch_access_token,
                           ut_twitch.refresh_token AS twitch_refresh_token,
                           ut_twitch.token_type    AS twitch_token_type,
                           ut_twitch.expires_in    AS twitch_expires_in,
                           ut_twitch.scope         AS twitch_scope,
                           ut_twitch.created_at    AS twitch_oauth_created_at,
                           ut_twitch.updated_at    AS twitch_oauth_updated_at,
                           ut_twitch.deleted_at    AS twitch_oauth_deleted_at,

                           -- Kick OAuth token
                           ut_kick.id              AS kick_oauth_id,
                           ut_kick.platform        AS kick_platform,
                           ut_kick.access_token    AS kick_access_token,
                           ut_kick.refresh_token   AS kick_refresh_token,
                           ut_kick.token_type      AS kick_token_type,
                           ut_kick.expires_in      AS kick_expires_in,
                           ut_kick.scope           AS kick_scope,
                           ut_kick.created_at      AS kick_oauth_created_at,
                           ut_kick.updated_at      AS kick_oauth_updated_at,
                           ut_kick.deleted_at      AS kick_oauth_deleted_at,

                           (
                                SELECT array_agg(DISTINCT ut.platform)
                                FROM user_oauth_tokens ut
                                WHERE ut.user_id = u.id
                                AND ut.platform NOT IN ('twitch', 'kick')
                            ) AS social_media_platforms
                    FROM users u
                             LEFT JOIN twitch_account ta
                                       ON u.id = ta.user_id
                             LEFT JOIN kick_account ka
                                       ON u.id = ka.user_id
                             LEFT JOIN user_oauth_tokens ut_twitch
                                       ON u.id = ut_twitch.user_id
                                           AND ut_twitch.platform = 'twitch'
                             LEFT JOIN user_oauth_tokens ut_kick
                                       ON u.id = ut_kick.user_id
                                           AND ut_kick.platform = 'kick'
                    """

SELECT_USER_BY_ID_QUERY = SELECT_USER_QUERY + "WHERE u.id = %s AND u.deleted_at IS NULL"
SELECT_USER_BY_EMAIL_QUERY = (
    SELECT_USER_QUERY + "WHERE u.email = %s AND u.deleted_at IS NULL"
)
SELECT_USER_BY_TWITCH_USERNAME_QUERY = (
    SELECT_USER_QUERY + "WHERE ta.twitch_username = %s AND u.deleted_at IS NULL"
)
SELECT_USER_BY_TWITCH_ID_QUERY = (
    SELECT_USER_QUERY + "WHERE ta.twitch_id = %s AND u.deleted_at IS NULL"
)
SELECT_USER_BY_KICK_USERNAME_QUERY = (
    SELECT_USER_QUERY + "WHERE ka.kick_username = %s AND u.deleted_at IS NULL"
)
SELECT_USER_BY_KICK_ID_QUERY = (
    SELECT_USER_QUERY + "WHERE ka.kick_id = %s AND u.deleted_at IS NULL"
)
SELECT_OAUTH_TOKEN_QUERY = """
                           SELECT id,
                                  user_id,
                                  platform,
                                  access_token,
                                  refresh_token,
                                  token_type,
                                  expires_in,
                                  scope,
                                  created_at,
                                  updated_at,
                                  deleted_at
                           FROM user_oauth_tokens
                           """

SELECT_TWITCH_ACCOUNT_QUERY = """
                              SELECT id,
                                     user_id,
                                     twitch_id,
                                     twitch_username,
                                     profile_link,
                                     profile_picture,
                                     created_at,
                                     updated_at,
                                     deleted_at
                              FROM twitch_account \
                              """

SELECT_KICK_ACCOUNT_QUERY = """
                            SELECT id,
                                   user_id,
                                   kick_id,
                                   kick_username,
                                   profile_link,
                                   profile_picture,
                                   created_at,
                                   updated_at,
                                   deleted_at
                            FROM kick_account \
                            """

logger = get_logger(__name__)


class UserNotFoundError(Exception):
    """Exception raised when a user is not found."""

    def __init__(self, message="User not found"):
        self.message = message
        super().__init__(self.message)


class UserAlreadyExistsError(Exception):
    """Exception raised when trying to create a user that already exists."""

    def __init__(self, message="User already exists"):
        self.message = message
        super().__init__(self.message)


class UserDAO:
    """Data Access Object for users in PostgreSQL."""

    def __init__(self, db: Db | None = None, crypto: Crypto | None = None):
        """Initialize with an existing DB connection or create a new one."""
        self.db = db if db is not None else Db(DBConfig())
        self.crypto = (
            crypto if crypto is not None else Crypto(config=CryptoConfig())  # type: ignore[missing-argument]
        )

    async def _upsert_owned_account(
        self,
        *,
        table: str,
        columns: list[str],
        values: tuple[Any, ...],
        conflict_column: str,
        update_columns: list[str],
    ) -> int:
        """Upsert an account row only when the conflicting provider id belongs to the same user."""

        placeholders = ", ".join(["%s"] * len(columns))
        columns_formatted = ", ".join(columns)
        update_clause = ", ".join(
            f"{column} = EXCLUDED.{column}" for column in update_columns
        )
        query = f"""
            INSERT INTO {table} ({columns_formatted})
            VALUES ({placeholders})
            ON CONFLICT ({conflict_column})
            DO UPDATE SET {update_clause}
            WHERE {table}.user_id = EXCLUDED.user_id
            RETURNING id;
        """

        async with await self.db.get_connection() as conn:
            lock = conn._get_lock()
            async with lock:
                async with conn.cursor() as cur:
                    await cur.execute(query.encode(), values)
                    result = await cur.fetchone()
                    if not result:
                        raise UniqueConstraintError(
                            f"{table} is already linked to another user."
                        )

            await conn.commit()

        return int(result[0])

    async def create_user(self, user: UserCreate) -> Optional[User]:
        """Create a new user."""
        with span(
            logger,
            "create_user",
            {"email": str(user.email)},
            log_entry=False,
            log_exit=False,
        ):
            try:
                # Check if user already exists
                try:
                    await self.get_user_by_email(user.email.__str__())
                    logger.warning(
                        f"User with email {user.email} already exists",
                        extra={"email": str(user.email)},
                    )
                    raise UserAlreadyExistsError(
                        f"User with email {user.email} already exists."
                    )
                except UserNotFoundError:
                    # This is the expected path when creating a new user
                    pass

                user_dict = user.dict()
                columns = []
                values = []
                placeholders = []

                for key, value in user_dict.items():
                    if value is not None:
                        columns.append(key)
                        values.append(value)
                        placeholders.append("%s")

                result = await self.db.insert(
                    table="users", columns=columns, values=tuple(values)
                )

                if result:
                    new_user = User(
                        id=result,
                        created_at=datetime.datetime.now(),
                        updated_at=datetime.datetime.now(),
                        **user_dict,
                    )
                    logger.info(
                        f"Created new user with ID {result}",
                        extra={"user_id": result, "email": str(user.email)},
                    )
                    return new_user

                logger.error(
                    f"Failed to create user with email {user.email}",
                    extra={"email": str(user.email)},
                )
                return None
            except Exception as e:
                if not isinstance(e, UserAlreadyExistsError):
                    logger.error(
                        f"Error creating user: {e}",
                        extra={"email": str(user.email), "error": str(e)},
                    )
                raise

    async def get_user_by_id(self, user_id: int) -> User:
        """Get a user by their ID."""
        with span(
            logger,
            "get_user_by_id",
            {"user_id": user_id},
            log_entry=False,
            log_exit=False,
        ):
            try:
                row = await self.db.fetch_one(SELECT_USER_BY_ID_QUERY, (user_id,))
                if not row:
                    logger.warning(
                        f"User with ID {user_id} not found", extra={"user_id": user_id}
                    )
                    raise UserNotFoundError(f"User with ID {user_id} not found.")

                user = self._row_to_user(row)
                logger.debug(
                    f"Found user with ID {user_id}",
                    extra={"user_id": user_id, "email": str(user.email)},
                )
                return user
            except Exception as e:
                if not isinstance(e, UserNotFoundError):
                    logger.error(
                        f"Error retrieving user with ID {user_id}: {e}",
                        extra={"user_id": user_id, "error": str(e)},
                    )
                raise

    async def get_user_by_email(self, email: str, **kwargs) -> User:
        """Get a user by their email."""
        with span(
            logger,
            "get_user_by_email",
            {"email": email, "only_user": kwargs.get("only_user", False)},
            log_entry=False,
            log_exit=False,
        ):
            try:
                if kwargs.get("only_user", False):
                    row = await self.db.fetch_one(
                        SELECT_ONLY_USER_QUERY + "WHERE email = %s", (email,)
                    )
                    if not row:
                        logger.warning(
                            f"User with email {email} not found", extra={"email": email}
                        )
                        raise UserNotFoundError(f"User with email {email} not found.")

                    user = self._row_to_user(row)
                    logger.debug(
                        f"Found user with email {email} (only user data)",
                        extra={"email": email, "user_id": user.id},
                    )
                    return user

                row = await self.db.fetch_one(SELECT_USER_BY_EMAIL_QUERY, (email,))
                if not row:
                    logger.warning(
                        f"User with email {email} not found", extra={"email": email}
                    )
                    raise UserNotFoundError(f"User with email {email} not found.")

                user = self._row_to_user(row)
                logger.debug(
                    f"Found user with email {email}",
                    extra={"email": email, "user_id": user.id},
                )
                return user
            except Exception as e:
                if not isinstance(e, UserNotFoundError):
                    logger.error(
                        f"Error retrieving user with email {email}: {e}",
                        extra={"email": email, "error": str(e)},
                    )
                raise

    async def get_user_by_twitch_username(self, username: str) -> User:
        """Get a user by their Twitch username."""
        with span(
            logger,
            "get_user_by_twitch_username",
            {"twitch_username": username},
            log_entry=False,
            log_exit=False,
        ):
            try:
                normalized_username = username.lower() if username else ""
                row = await self.db.fetch_one(
                    SELECT_USER_BY_TWITCH_USERNAME_QUERY, (normalized_username,)
                )
                if not row:
                    logger.warning(
                        f"User with Twitch username {username} not found",
                        extra={"twitch_username": username},
                    )
                    raise UserNotFoundError(
                        f"User with Twitch username {username} not found."
                    )

                user = self._row_to_user(row)
                logger.debug(
                    f"Found user with Twitch username {username}",
                    extra={"twitch_username": username, "user_id": user.id},
                )
                return user
            except Exception as e:
                if not isinstance(e, UserNotFoundError):
                    logger.error(
                        f"Error retrieving user with Twitch username {username}: {e}",
                        extra={"twitch_username": username, "error": str(e)},
                    )
                raise

    async def get_user_by_twitch_id(self, twitch_id: str) -> User:
        """Get a user by their Twitch broadcast ID."""
        with span(
            logger,
            "get_user_by_twitch_id",
            {"twitch_id": twitch_id},
            log_entry=False,
            log_exit=False,
        ):
            try:
                row = await self.db.fetch_one(
                    SELECT_USER_BY_TWITCH_ID_QUERY, (twitch_id,)
                )
                if not row:
                    logger.warning(
                        f"User with Twitch ID {twitch_id} not found",
                        extra={"twitch_id": twitch_id},
                    )
                    raise UserNotFoundError(
                        f"User with Twitch ID {twitch_id} not found."
                    )

                user = self._row_to_user(row)
                logger.debug(
                    f"Found user with Twitch ID {twitch_id}",
                    extra={"twitch_id": twitch_id, "user_id": user.id},
                )
                return user
            except Exception as e:
                if not isinstance(e, UserNotFoundError):
                    logger.error(
                        f"Error retrieving user with Twitch ID {twitch_id}: {e}",
                        extra={"twitch_id": twitch_id, "error": str(e)},
                    )
                raise

    async def get_user_by_kick_username(self, username: str) -> User:
        """Get a user by their Kick username."""
        with span(
            logger,
            "get_user_by_kick_username",
            {"kick_username": username},
            log_entry=False,
            log_exit=False,
        ):
            try:
                normalized_username = username.lower() if username else ""
                row = await self.db.fetch_one(
                    SELECT_USER_BY_KICK_USERNAME_QUERY, (normalized_username,)
                )
                if not row:
                    logger.warning(
                        f"User with Kick username {username} not found",
                        extra={"kick_username": username},
                    )
                    raise UserNotFoundError(
                        f"User with Kick username {username} not found."
                    )

                user = self._row_to_user(row)
                logger.debug(
                    f"Found user with Kick username {username}",
                    extra={"kick_username": username, "user_id": user.id},
                )
                return user
            except Exception as e:
                if not isinstance(e, UserNotFoundError):
                    logger.error(
                        f"Error retrieving user with Kick username {username}: {e}",
                        extra={"kick_username": username, "error": str(e)},
                    )
                raise

    async def get_user_by_kick_id(self, kick_id: str) -> User:
        """Get a user by their Kick broadcast ID."""
        with span(
            logger,
            "get_user_by_kick_id",
            {"kick_id": kick_id},
            log_entry=False,
            log_exit=False,
        ):
            try:
                row = await self.db.fetch_one(SELECT_USER_BY_KICK_ID_QUERY, (kick_id,))
                if not row:
                    logger.warning(
                        f"User with Kick ID {kick_id} not found",
                        extra={"kick_id": kick_id},
                    )
                    raise UserNotFoundError(f"User with Kick ID {kick_id} not found.")

                user = self._row_to_user(row)
                logger.debug(
                    f"Found user with Kick ID {kick_id}",
                    extra={"kick_id": kick_id, "user_id": user.id},
                )
                return user
            except Exception as e:
                if not isinstance(e, UserNotFoundError):
                    logger.error(
                        f"Error retrieving user with Kick ID {kick_id}: {e}",
                        extra={"kick_id": kick_id, "error": str(e)},
                    )
                raise

    async def update_user_by_email(self, email: str, update: UserUpdate) -> None:
        """Update a user's details."""
        with span(logger, "update_user_by_email", {"email": email}):
            update_dict = update.dict(exclude_none=True)
            if not update_dict:
                logger.debug(
                    f"No updates provided for user email {email}, skipping",
                    extra={"email": email},
                )
                return

            set_clauses = []
            values = []

            for key, value in update_dict.items():
                set_clauses.append(key)
                values.append(value)

            try:
                await self.db.update(
                    table="users",
                    set_columns=set_clauses,
                    set_values=values,
                    where_clause="email = %s",
                    where_params=(email,),
                )
                logger.info(
                    f"Updated user email {email} with fields: {list(update_dict.keys())}",
                    extra={"email": email, "fields": list(update_dict.keys())},
                )
            except Exception as e:
                logger.error(
                    f"Error updating user email {email}: {e}",
                    extra={"email": email, "error": str(e)},
                )
                raise

    async def update_user(self, user_id: int, update: UserUpdate) -> None:
        """Update a user's details."""
        with span(
            logger, "update_user", {"user_id": user_id}, log_entry=False, log_exit=False
        ):
            update_dict = update.dict(exclude_none=True)
            if not update_dict:
                logger.debug(
                    f"No updates provided for user ID {user_id}, skipping",
                    extra={"user_id": user_id},
                )
                return

            set_clauses = []
            values = []

            for key, value in update_dict.items():
                set_clauses.append(key)
                values.append(value)

            try:
                await self.db.update(
                    table="users",
                    set_columns=set_clauses,
                    set_values=values,
                    where_clause="id = %s",
                    where_params=(user_id,),
                )
                logger.info(
                    f"Updated user ID {user_id} with fields: {list(update_dict.keys())}",
                    extra={"user_id": user_id, "fields": list(update_dict.keys())},
                )
            except Exception as e:
                logger.error(
                    f"Error updating user ID {user_id}: {e}",
                    extra={"user_id": user_id, "error": str(e)},
                )
                raise

    async def delete_user(self, user_id: int) -> None:
        """Delete a user."""
        await self.db.delete(
            table="users", where_clause="id = %s", where_params=(user_id,)
        )

    async def get_twitch_account(self, user_id: int) -> Optional[TwitchAccount]:
        """
        Fetch a user's Twitch account details.
        """
        row = await self.db.fetch_one(
            SELECT_TWITCH_ACCOUNT_QUERY + "WHERE user_id = %s",
            (user_id,),
        )

        if not row:
            raise Exception("Twitch account not found for user.")

        return TwitchAccount(
            id=row[0],
            user_id=row[1],
            twitch_id=row[2],
            twitch_username=row[3],
            profile_link=row[4],
            profile_picture=row[5],
            created_at=row[6],
            updated_at=row[7],
            deleted_at=row[8],
        )

    async def get_kick_account(self, user_id: int) -> Optional[KickAccount]:
        """
        Fetch a user's Kick account details.
        """
        row = await self.db.fetch_one(
            SELECT_KICK_ACCOUNT_QUERY + "WHERE user_id = %s",
            (user_id,),
        )

        if not row:
            raise Exception("Kick account not found for user.")

        return KickAccount(
            id=row[0],
            user_id=row[1],
            kick_id=row[2],
            kick_username=row[3],
            profile_link=row[4],
            profile_picture=row[5],
            created_at=row[6],
            updated_at=row[7],
            deleted_at=row[8],
        )

    async def get_twitch_account_by_username(
        self, username: str
    ) -> Optional[TwitchAccount]:
        """
        Fetch a user's Twitch account details by username.
        """
        normalized_username = username.lower() if username else ""
        row = await self.db.fetch_one(
            SELECT_TWITCH_ACCOUNT_QUERY + "WHERE twitch_username = %s",
            (normalized_username,),
        )

        if not row:
            raise Exception("Twitch account not found for user.")

        return TwitchAccount(
            id=row[0],
            user_id=row[1],
            twitch_id=row[2],
            twitch_username=row[3],
            profile_link=row[4],
            profile_picture=row[5],
            created_at=row[6],
            updated_at=row[7],
            deleted_at=row[8],
        )

    async def get_kick_account_by_username(
        self, username: str
    ) -> Optional[KickAccount]:
        """
        Fetch a user's Kick account details by username.
        """
        normalized_username = username.lower() if username else ""
        row = await self.db.fetch_one(
            SELECT_KICK_ACCOUNT_QUERY + "WHERE kick_username = %s",
            (normalized_username,),
        )

        if not row:
            raise Exception("Kick account not found for user.")

        return KickAccount(
            id=row[0],
            user_id=row[1],
            kick_id=row[2],
            kick_username=row[3],
            profile_link=row[4],
            profile_picture=row[5],
            created_at=row[6],
            updated_at=row[7],
            deleted_at=row[8],
        )

    async def create_twitch_account(
        self, account: TwitchAccountCreate
    ) -> TwitchAccount:
        """
        Insert a new Twitch account for the user.
        """
        res = await self.db.insert(
            table="twitch_account",
            columns=[
                "user_id",
                "twitch_id",
                "twitch_username",
                "profile_link",
                "profile_picture",
                "authorized_at",
                "oauth_token_id",
            ],
            values=(
                account.user_id,
                account.twitch_id,
                account.twitch_username.lower(),
                account.profile_link,
                account.profile_picture,
                account.authorized_at,
                account.oauth_token_id,
            ),
        )
        if not res:
            raise Exception("Failed to create Twitch account for user.")
        return TwitchAccount(
            id=res,
            user_id=account.user_id,
            twitch_id=account.twitch_id,
            twitch_username=account.twitch_username.lower(),
            profile_link=account.profile_link,
            profile_picture=account.profile_picture,
            authorized_at=account.authorized_at,
            oauth_token_id=account.oauth_token_id,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

    async def create_kick_account(self, account: KickAccountCreate) -> KickAccount:
        """
        Insert a new Kick account for the user.
        """
        res = await self.db.insert(
            table="kick_account",
            columns=[
                "user_id",
                "kick_id",
                "kick_username",
                "profile_link",
                "profile_picture",
            ],
            values=(
                account.user_id,
                account.kick_id,
                account.kick_username.lower(),
                account.profile_link,
                account.profile_picture,
            ),
        )
        if not res:
            raise Exception("Failed to create Kick account for user.")
        return KickAccount(
            id=res,
            user_id=account.user_id,
            kick_id=account.kick_id,
            kick_username=account.kick_username.lower(),
            profile_link=account.profile_link,
            profile_picture=account.profile_picture,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

    async def upsert_twitch_account(
        self, account: TwitchAccountCreate
    ) -> TwitchAccount:
        """
        Insert a new Twitch account for the user or update if it already exists.
        """
        normalized_twitch_username = account.twitch_username.lower()
        res = await self._upsert_owned_account(
            table="twitch_account",
            columns=[
                "user_id",
                "twitch_id",
                "twitch_username",
                "profile_link",
                "profile_picture",
                "authorized_at",
                "oauth_token_id",
            ],
            values=(
                account.user_id,
                account.twitch_id,
                normalized_twitch_username,
                account.profile_link,
                account.profile_picture,
                account.authorized_at,
                account.oauth_token_id,
            ),
            conflict_column="twitch_id",
            update_columns=[
                "twitch_username",
                "profile_link",
                "profile_picture",
                "authorized_at",
                "oauth_token_id",
            ],
        )
        if not res:
            raise Exception("Failed to create or update Twitch account for user.")
        return TwitchAccount(
            id=res,
            user_id=account.user_id,
            twitch_id=account.twitch_id,
            twitch_username=normalized_twitch_username,
            profile_link=account.profile_link,
            profile_picture=account.profile_picture,
            authorized_at=account.authorized_at,
            oauth_token_id=account.oauth_token_id,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

    async def upsert_social_media_account(
        self, account: SocialMediaAccountCreate
    ) -> SocialMediaAccount:
        """
        Insert a new Social Media account for the user or update if it already exists.
        """

        res = await self.db.upsert(
            table="social_media_account",
            columns=["user_id", "platform", "oauth_token_id", "username"],
            values=(
                account.user_id,
                account.platform.value,
                account.oauth_token_id,
                account.username,
            ),
            conflict_columns=["user_id", "platform"],
            update_columns=["oauth_token_id", "username"],
        )
        if not res:
            raise Exception("Failed to create or update Social Media account for user.")
        return SocialMediaAccount(
            id=res,
            user_id=account.user_id,
            platform=account.platform,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
            username=account.username,
        )

    async def upsert_kick_account(self, account: KickAccountCreate) -> KickAccount:
        """
        Insert a new Kick account for the user or update if it already exists.
        """
        normalized_kick_username = account.kick_username.lower()
        res = await self._upsert_owned_account(
            table="kick_account",
            columns=[
                "user_id",
                "kick_id",
                "kick_username",
                "profile_link",
                "profile_picture",
            ],
            values=(
                account.user_id,
                account.kick_id,
                normalized_kick_username,
                account.profile_link,
                account.profile_picture,
            ),
            conflict_column="kick_id",
            update_columns=[
                "profile_link",
                "profile_picture",
                "kick_username",
            ],
        )
        if not res:
            raise Exception("Failed to create or update Kick account for user.")
        return KickAccount(
            id=res,
            user_id=account.user_id,
            kick_id=account.kick_id,
            kick_username=normalized_kick_username,
            profile_link=account.profile_link,
            profile_picture=account.profile_picture,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

    async def update_twitch_account(
        self, user_id: int, account: TwitchAccountUpdate
    ) -> None:
        """
        Update (overwrite) an existing Twitch account row.
        """
        account.twitch_username = (
            account.twitch_username.lower() if account.twitch_username else None
        )
        update_dict = account.dict(exclude_none=True)
        if not update_dict:
            return

        set_clauses = []
        values = []

        for key, value in update_dict.items():
            set_clauses.append(key)
            values.append(value)

        await self.db.update(
            table="twitch_account",
            set_columns=set_clauses,
            set_values=values,
            where_clause="user_id = %s",
            where_params=(user_id,),
        )

    async def update_kick_account(
        self, user_id: int, account: KickAccountUpdate
    ) -> None:
        """
        Update (overwrite) an existing Kick account row.
        """
        account.kick_username = (
            account.kick_username.lower() if account.kick_username else None
        )
        update_dict = account.dict(exclude_none=True)
        if not update_dict:
            return

        set_clauses = []
        values = []

        for key, value in update_dict.items():
            set_clauses.append(key)
            values.append(value)

        await self.db.update(
            table="kick_account",
            set_columns=set_clauses,
            set_values=values,
            where_clause="user_id = %s",
            where_params=(user_id,),
        )

    async def delete_twitch_account(self, user_id: int) -> None:
        """
        Delete a user's Twitch account.
        """
        await self.db.delete(
            table="twitch_account",
            where_clause="user_id = %s",
            where_params=(user_id,),
        )

    async def delete_kick_account(self, user_id: int) -> None:
        """
        Delete a user's Kick account.
        """
        await self.db.delete(
            table="kick_account",
            where_clause="user_id = %s",
            where_params=(user_id,),
        )

    async def list_all_twitch_accounts(self) -> list[tuple[int, str]]:
        """List all active Twitch accounts (user_id, twitch_id).

        Returns:
            List of tuples: (user_id, twitch_id)
        """
        rows = await self.db.fetch_all(
            "SELECT user_id, twitch_id FROM twitch_account WHERE deleted_at IS NULL",
            (),
        )
        return [(row[0], str(row[1])) for row in (rows or [])]

    async def list_twitch_accounts_with_last_login(
        self,
    ) -> list[tuple[int, str, datetime.datetime, str]]:
        """
        List all active Twitch accounts with last_login and tier metadata.

        Returns:
            List of tuples: (user_id, twitch_id, last_login, tier)
        """
        rows = await self.db.fetch_all(
            """
            SELECT ta.user_id, ta.twitch_id, u.last_login, u.tier
            FROM twitch_account ta
            JOIN users u ON u.id = ta.user_id
            WHERE ta.deleted_at IS NULL AND u.deleted_at IS NULL
            """,
            (),
        )
        return [(row[0], str(row[1]), row[2], row[3]) for row in (rows or [])]

    async def list_all_kick_accounts(self) -> list[tuple[int, str]]:
        """List all active Kick accounts (user_id, kick_id).

        Returns:
            List of tuples: (user_id, kick_id)
        """
        rows = await self.db.fetch_all(
            "SELECT user_id, kick_id FROM kick_account WHERE deleted_at IS NULL",
            (),
        )
        return [(row[0], str(row[1])) for row in (rows or [])]

    async def list_kick_accounts_with_last_login(
        self,
    ) -> list[tuple[int, str, datetime.datetime, str]]:
        """
        List all active Kick accounts with last_login and tier metadata.

        Returns:
            List of tuples: (user_id, kick_id, last_login, tier)
        """
        rows = await self.db.fetch_all(
            """
            SELECT ka.user_id, ka.kick_id, u.last_login, u.tier
            FROM kick_account ka
            JOIN users u ON u.id = ka.user_id
            WHERE ka.deleted_at IS NULL AND u.deleted_at IS NULL
            """,
            (),
        )
        return [(row[0], str(row[1]), row[2], row[3]) for row in (rows or [])]

    async def get_user_oauth_token(
        self, user_id: int, platform: Platform
    ) -> UserAccessToken:
        """
        Fetch a user's OAuth token for the given platform.
        """
        row = await self.db.fetch_one(
            SELECT_OAUTH_TOKEN_QUERY + "WHERE user_id = %s AND platform = %s",
            (user_id, platform.value),
        )

        if not row:
            raise Exception(
                f"{platform.value.capitalize()} OAuth token not found for user."
            )

        return self._row_to_oauth_token(row)

    async def get_valid_twitch_oauth_token(
        self, user_id: int, twitch_api, force_refresh: bool = False
    ) -> UserAccessToken:
        """Return a valid Twitch OAuth token for a user.

        When ``force_refresh`` is True, refreshes even if token isn't time-expired.
        """
        try:
            user_token = await self.get_user_oauth_token(user_id, Platform.TWITCH)
        except Exception as exc:
            raise Exception(
                f"Twitch OAuth token not found for user {user_id}: {exc}"
            ) from exc

        try:
            if force_refresh:
                return await self.refresh_twitch_token(user_id, user_token, twitch_api)
            return await self.refresh_twitch_token_if_needed(
                user_id, user_token, twitch_api
            )
        except Exception as exc:
            raise Exception(
                f"Failed to obtain valid Twitch OAuth token for user {user_id}: {exc}"
            ) from exc

    async def create_user_oauth_token(
        self, token: UserAccessTokenCreate
    ) -> UserAccessToken:
        """
        Insert a new OAuth token (one row per user & platform – unique enforced).
        """
        enc_access = self.crypto.encrypt(token.access_token.encode())
        enc_refresh = (
            self.crypto.encrypt(token.refresh_token.encode())
            if token.refresh_token
            else None
        )

        res = await self.db.insert(
            table="user_oauth_tokens",
            columns=[
                "user_id",
                "platform",
                "access_token",
                "refresh_token",
                "token_type",
                "expires_in",
                "scope",
            ],
            values=(
                token.user_id,
                token.platform.value,
                enc_access,
                enc_refresh,
                token.token_type,
                token.expires_in,
                ",".join(token.scope or []),
            ),
        )
        if not res:
            raise Exception(
                f"Failed to create {token.platform.value} OAuth token for user."
            )

        return UserAccessToken(
            id=res,
            user_id=token.user_id,
            platform=token.platform,
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            token_type=token.token_type,
            expires_in=token.expires_in,
            scope=token.scope,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

    async def upsert_user_oauth_token(
        self, token: UserAccessTokenCreate
    ) -> UserAccessToken:
        """
        Insert a new OAuth token (one row per user & platform – unique enforced).
        """
        enc_access = self.crypto.encrypt(token.access_token.encode())
        enc_refresh = (
            self.crypto.encrypt(token.refresh_token.encode())
            if token.refresh_token
            else None
        )

        res = await self.db.upsert(
            table="user_oauth_tokens",
            columns=[
                "user_id",
                "platform",
                "access_token",
                "refresh_token",
                "token_type",
                "expires_in",
                "scope",
                "updated_at",
            ],
            values=(
                token.user_id,
                token.platform.value,
                enc_access,
                enc_refresh,
                token.token_type,
                token.expires_in,
                ",".join(token.scope or []),
                datetime.datetime.utcnow(),
            ),
            conflict_columns=["user_id", "platform"],
            update_columns=[
                "access_token",
                "refresh_token",
                "token_type",
                "expires_in",
                "scope",
                "updated_at",
            ],
        )
        if not res:
            raise Exception(
                f"Failed to create {token.platform.value} OAuth token for user."
            )

        return UserAccessToken(
            id=res,
            user_id=token.user_id,
            platform=token.platform,
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            token_type=token.token_type,
            expires_in=token.expires_in,
            scope=token.scope,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
        )

    async def update_user_oauth_token(
        self, user_id: int, token: UserAccessTokenUpdate
    ) -> None:
        """
        Update (overwrite) an existing OAuth token row.
        """
        if token.access_token is None:
            raise ValueError("OAuth token update requires access_token")
        if token.platform is None:
            raise ValueError("OAuth token update requires platform")

        enc_access = self.crypto.encrypt(token.access_token.encode())
        enc_refresh = (
            self.crypto.encrypt(token.refresh_token.encode())
            if token.refresh_token
            else None
        )

        await self.db.update(
            table="user_oauth_tokens",
            set_columns=[
                "access_token",
                "refresh_token",
                "token_type",
                "expires_in",
                "scope",
                "updated_at",
            ],
            set_values=[
                enc_access,
                enc_refresh,
                token.token_type,
                token.expires_in,
                ",".join(token.scope or []),
                "NOW()",
            ],
            where_clause="user_id = %s AND platform = %s",
            where_params=(user_id, token.platform.value),
        )

    async def delete_user_oauth_token(self, user_id: int, platform: Platform) -> None:
        """
        Delete a user's OAuth token for the given platform.
        """
        await self.db.delete(
            table="user_oauth_tokens",
            where_clause="user_id = %s AND platform = %s",
            where_params=(user_id, platform.value),
        )

    async def refresh_kick_token_if_needed(
        self, user_id: int, user_token: UserAccessToken, kick_api
    ) -> UserAccessToken:
        """
        Check if Kick token is expired and refresh if needed.

        Args:
            user_id: The user ID
            user_token: Current user access token
            kick_api: KickApi instance for refreshing tokens

        Returns:
            UserAccessToken: The current token if valid, or refreshed token

        Raises:
            ValueError: If no refresh token is available
            Exception: If token refresh fails
        """
        # Check if token is expired
        if not user_token.is_expired():
            return user_token

        if not user_token.refresh_token:
            raise ValueError(f"No refresh token available for user {user_id}")

        logger.info(
            "Kick token expired, refreshing",
            extra={"user_id": user_id, "platform": "kick"},
        )

        # Refresh the token
        token_data = await kick_api.refresh_user_token(user_token.refresh_token)

        # Update in database
        token_update = UserAccessTokenUpdate(
            access_token=token_data["access_token"],
            refresh_token=token_data.get("refresh_token"),
            expires_in=token_data["expires_in"],
            token_type=token_data["token_type"],
            scope=(
                token_data.get("scope", "").split(" ")
                if isinstance(token_data.get("scope"), str)
                else token_data.get("scope")
            )
            if token_data.get("scope")
            else None,
            platform=Platform.KICK,
        )
        await self.update_user_oauth_token(user_id, token_update)

        logger.info(
            "Successfully refreshed Kick token",
            extra={"user_id": user_id, "platform": "kick"},
        )

        # Return refreshed token from database
        return await self.get_user_oauth_token(user_id, Platform.KICK)

    async def refresh_twitch_token_if_needed(
        self, user_id: int, user_token: UserAccessToken, twitch_api
    ) -> UserAccessToken:
        """
        Check if Twitch token is expired and refresh if needed.

        Args:
            user_id: The user ID
            user_token: Current user access token
            twitch_api: TwitchApi instance for refreshing tokens

        Returns:
            UserAccessToken: The current token if valid, or refreshed token

        Raises:
            ValueError: If no refresh token is available
            Exception: If token refresh fails
        """
        # Check if token is expired
        if not user_token.is_expired():
            return user_token

        return await self.refresh_twitch_token(user_id, user_token, twitch_api)

    async def refresh_twitch_token(
        self, user_id: int, user_token: UserAccessToken, twitch_api
    ) -> UserAccessToken:
        """Refresh a Twitch user token regardless of current expiry state."""
        if not user_token.refresh_token:
            raise ValueError(f"No refresh token available for user {user_id}")

        logger.info(
            "Refreshing Twitch token",
            extra={"user_id": user_id, "platform": "twitch"},
        )

        # Refresh the token
        token_data = await twitch_api.refresh_user_token(user_token.refresh_token)

        # Update in database
        token_update = UserAccessTokenUpdate(
            access_token=token_data["access_token"],
            refresh_token=token_data.get("refresh_token"),
            expires_in=token_data["expires_in"],
            token_type=token_data["token_type"],
            scope=(
                token_data.get("scope", "").split(" ")
                if isinstance(token_data.get("scope"), str)
                else token_data.get("scope")
            )
            if token_data.get("scope")
            else None,
            platform=Platform.TWITCH,
        )
        await self.update_user_oauth_token(user_id, token_update)

        logger.info(
            "Successfully refreshed Twitch token",
            extra={"user_id": user_id, "platform": "twitch"},
        )

        # Return refreshed token from database
        return await self.get_user_oauth_token(user_id, Platform.TWITCH)

    async def get_name_by_user_id(self, user_id: int) -> str:
        """Return a display name for the user.

        Priority:
        1) Twitch username
        2) Kick username
        3) First name + last name (fallback)
        """
        query = """
            SELECT
                ta.twitch_username,
                ka.kick_username,
                u.first_name,
                u.last_name
            FROM users u
            LEFT JOIN twitch_account ta ON u.id = ta.user_id AND ta.deleted_at IS NULL
            LEFT JOIN kick_account   ka ON u.id = ka.user_id AND ka.deleted_at IS NULL
            WHERE u.id = %s AND u.deleted_at IS NULL
            LIMIT 1
        """

        row = await self.db.fetch_one(query, (user_id,))
        if not row:
            raise UserNotFoundError(f"User with ID {user_id} not found.")

        twitch_username, kick_username, first_name, last_name = row

        if isinstance(twitch_username, str) and twitch_username.strip():
            return twitch_username.strip()

        if isinstance(kick_username, str) and kick_username.strip():
            return kick_username.strip()

        fn = first_name.strip() if isinstance(first_name, str) else ""
        ln = last_name.strip() if isinstance(last_name, str) else ""
        full = f"{fn} {ln}".strip()

        return full if full else ""

    @staticmethod
    def _row_to_twitch_account(row) -> TwitchAccount | None:
        """Convert a database row to a TwitchAccount model."""
        if not row:
            return None
        if len(row) < 9:
            raise ValueError(
                "Row does not contain enough data to create a TwitchAccount object."
            )
        account_data = row[:9]
        return TwitchAccount(
            id=account_data[0],
            user_id=account_data[1],
            twitch_id=account_data[2],
            twitch_username=account_data[3],
            profile_link=account_data[4],
            profile_picture=account_data[5],
            created_at=account_data[6],
            updated_at=account_data[7],
            deleted_at=account_data[8],
        )

    @staticmethod
    def _row_to_kick_account(row) -> KickAccount | None:
        """Convert a database row to a KickAccount model."""
        if not row:
            return None
        if len(row) < 9:
            raise ValueError(
                "Row does not contain enough data to create a KickAccount object."
            )
        account_data = row[:9]
        return KickAccount(
            id=account_data[0],
            user_id=account_data[1],
            kick_id=account_data[2],
            kick_username=account_data[3],
            profile_link=account_data[4],
            profile_picture=account_data[5],
            created_at=account_data[6],
            updated_at=account_data[7],
            deleted_at=account_data[8],
        )

    def _row_to_oauth_token(self, row: tuple[Any, ...]) -> UserAccessToken:
        """Convert a database row to an AccessToken model."""
        if not row:
            raise ValueError("Row is empty; cannot create an AccessToken object.")
        if len(row) < 11:
            raise ValueError(
                "Row does not contain enough data to create an AccessToken object."
            )
        token_data = row[:11]
        platform_raw = token_data[2]
        if isinstance(platform_raw, Platform):
            platform = platform_raw
        elif isinstance(platform_raw, str):
            platform = Platform(platform_raw.lower())
        else:
            raise ValueError("OAuth token platform is missing or invalid.")

        access_token = self._decrypt_token_value(token_data[3])
        if access_token is None:
            raise ValueError("OAuth token access_token is missing.")

        token_type_raw = token_data[5]
        token_type = str(token_type_raw) if token_type_raw is not None else ""
        expires_in_raw = token_data[6]
        expires_in = int(expires_in_raw) if expires_in_raw is not None else 0
        scope_raw = token_data[7]
        scope = (
            scope_raw.split(",") if isinstance(scope_raw, str) and scope_raw else None
        )
        return UserAccessToken(
            id=token_data[0],
            user_id=token_data[1],
            platform=platform,
            access_token=access_token,
            refresh_token=self._decrypt_token_value(token_data[4]),
            token_type=token_type,
            expires_in=expires_in,
            scope=scope,
            created_at=token_data[8],
            updated_at=token_data[9],
            deleted_at=token_data[10],
        )

    def _row_to_user(self, row: tuple[Any, ...]) -> User:
        """Convert a database row to a User model."""
        if not row:
            raise ValueError("Row is empty; cannot create a User object.")
        if len(row) < USER_COLUMNS_LEN:
            raise ValueError(
                "Row does not contain enough data to create a User object."
            )

        # Extract data from the row based on the new query structure
        user_data = row[:USER_COLUMNS_LEN]
        twitch_account_data = row[
            USER_COLUMNS_LEN : USER_COLUMNS_LEN + TWITCH_ACCOUNT_COLUMNS_LEN
        ]
        kick_account_data = row[
            USER_COLUMNS_LEN + TWITCH_ACCOUNT_COLUMNS_LEN : USER_COLUMNS_LEN
            + TWITCH_ACCOUNT_COLUMNS_LEN
            + KICK_ACCOUNT_COLUMNS_LEN
        ]
        twitch_oauth_data = row[
            USER_COLUMNS_LEN
            + TWITCH_ACCOUNT_COLUMNS_LEN
            + KICK_ACCOUNT_COLUMNS_LEN : USER_COLUMNS_LEN
            + TWITCH_ACCOUNT_COLUMNS_LEN
            + KICK_ACCOUNT_COLUMNS_LEN
            + TWITCH_ACCOUNT_OAUTH_COLUMNS_LEN
        ]
        kick_oauth_data = row[
            USER_COLUMNS_LEN
            + TWITCH_ACCOUNT_COLUMNS_LEN
            + KICK_ACCOUNT_COLUMNS_LEN
            + TWITCH_ACCOUNT_OAUTH_COLUMNS_LEN : USER_COLUMNS_LEN
            + TWITCH_ACCOUNT_COLUMNS_LEN
            + KICK_ACCOUNT_COLUMNS_LEN
            + TWITCH_ACCOUNT_OAUTH_COLUMNS_LEN
            + KICK_ACCOUNT_OAUTH_COLUMNS_LEN
        ]

        social_media_platforms = row[
            USER_COLUMNS_LEN
            + TWITCH_ACCOUNT_COLUMNS_LEN
            + KICK_ACCOUNT_COLUMNS_LEN
            + TWITCH_ACCOUNT_OAUTH_COLUMNS_LEN
            + KICK_ACCOUNT_OAUTH_COLUMNS_LEN
        ]

        # Create the user object with basic user data
        user = User(
            id=user_data[0],
            first_promoter_id=user_data[1],
            first_name=user_data[2],
            last_name=user_data[3],
            email=user_data[4],
            phone=user_data[5],
            tier=user_data[6],
            product=user_data[7],
            created_at=user_data[8],
            updated_at=user_data[9],
            deleted_at=user_data[10],
            last_login=user_data[11],
            customer_journey_completed_at=user_data[12],
            social_media_platforms=social_media_platforms,
        )

        # Handle Twitch account data
        # Check if all required fields for TwitchAccount are not None
        if (
            not any(twitch_account_data)
            or twitch_account_data[0] is None  # id
            or twitch_account_data[1] is None  # twitch_id
            or twitch_account_data[2] is None  # twitch_username
            or twitch_account_data[5] is None  # created_at
            or twitch_account_data[6] is None
        ):  # updated_at
            user.twitch_account = None
        else:
            user.twitch_account = TwitchAccount(
                id=twitch_account_data[0],
                user_id=user.id,
                twitch_id=twitch_account_data[1],
                twitch_username=twitch_account_data[2],
                profile_link=twitch_account_data[3],
                profile_picture=twitch_account_data[4],
                created_at=twitch_account_data[5],
                updated_at=twitch_account_data[6],
                deleted_at=twitch_account_data[7],
            )

            # Add Twitch OAuth token if available
            if any(twitch_oauth_data):
                twitch_access_token = self._decrypt_token_value(twitch_oauth_data[2])
                if twitch_access_token is None:
                    raise ValueError("Missing Twitch access token.")
                user.twitch_account.access_token = UserAccessToken(
                    id=twitch_oauth_data[0],
                    user_id=user.id,
                    platform=Platform.TWITCH,
                    access_token=twitch_access_token,
                    refresh_token=self._decrypt_token_value(twitch_oauth_data[3]),
                    token_type=twitch_oauth_data[4],
                    expires_in=twitch_oauth_data[5],
                    scope=(
                        twitch_oauth_data[6].split(",")
                        if isinstance(twitch_oauth_data[6], str)
                        else twitch_oauth_data[6]
                    )
                    if twitch_oauth_data[6]
                    else None,
                    created_at=twitch_oauth_data[7],
                    updated_at=twitch_oauth_data[8],
                    deleted_at=twitch_oauth_data[9],
                )
                logger.info(
                    f"Twitch OAuth token successfully decrypted for user {user.id}"
                )

        # Handle Kick account data
        # Check if all required fields for KickAccount are not None
        if (
            not any(kick_account_data)
            or kick_account_data[0] is None  # id
            or kick_account_data[1] is None  # kick_id
            or kick_account_data[2] is None  # kick_username
            or kick_account_data[5] is None  # created_at
            or kick_account_data[6] is None
        ):  # updated_at
            user.kick_account = None
        else:
            user.kick_account = KickAccount(
                id=kick_account_data[0],
                user_id=user.id,
                kick_id=kick_account_data[1],
                kick_username=kick_account_data[2],
                profile_link=kick_account_data[3],
                profile_picture=kick_account_data[4],
                created_at=kick_account_data[5],
                updated_at=kick_account_data[6],
                deleted_at=kick_account_data[7],
            )

            # Add Kick OAuth token if available
            if any(kick_oauth_data):
                kick_access_token = self._decrypt_token_value(kick_oauth_data[2])
                if kick_access_token is None:
                    raise ValueError("Missing Kick access token.")
                user.kick_account.access_token = UserAccessToken(
                    id=kick_oauth_data[0],
                    user_id=user.id,
                    platform=Platform.KICK,
                    access_token=kick_access_token,
                    refresh_token=self._decrypt_token_value(kick_oauth_data[3]),
                    token_type=kick_oauth_data[4],
                    expires_in=kick_oauth_data[5],
                    scope=(
                        kick_oauth_data[6].split(",")
                        if isinstance(kick_oauth_data[6], str)
                        else kick_oauth_data[6]
                    )
                    if kick_oauth_data[6]
                    else None,
                    created_at=kick_oauth_data[7],
                    updated_at=kick_oauth_data[8],
                    deleted_at=kick_oauth_data[9],
                )

        return user

    def _decrypt_token_value(self, token: Any) -> str | None:
        if token is None:
            return None
        token_bytes: bytes
        if isinstance(token, bytes):
            token_bytes = token
        elif isinstance(token, memoryview):
            token_bytes = token.tobytes()
        elif isinstance(token, str):
            token_bytes = token.encode()
        else:
            raise TypeError(f"Unsupported token payload type: {type(token)!r}")
        return self.crypto.decrypt(token_bytes).decode()

    def decrypt_token(self, token: Any) -> str:
        decrypted = self._decrypt_token_value(token)
        if decrypted is None:
            raise ValueError("Token payload cannot be None")
        return decrypted

    def encrypt_token(self, token: str) -> bytes:
        return self.crypto.encrypt(token.encode())



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, ConfigDict


class StreamBase(BaseModel):
    """Base model for stream data."""

    game_name: Optional[str] = None
    user_id: int
    stream_id: str
    stream_start_time: datetime
    stream_end_time: Optional[datetime] = None
    platform: Optional[str] = None  # New field to specify the platform


class StreamCreate(StreamBase):
    """Model for creating a new stream."""

    pass


class StreamUpdate(BaseModel):
    """Model for updating stream data."""

    game_name: Optional[str] = None
    stream_start_time: Optional[datetime] = None
    stream_end_time: Optional[datetime] = None
    platform: Optional[str] = None  # New field to specify the platform


class Stream(StreamBase):
    """Complete stream model including database ID and end time."""

    id: int

    model_config = ConfigDict(from_attributes=True)



import datetime
import uuid
from typing import Optional, Any

from common.database import Db, DBConfig
from common.logging import get_logger


logger = get_logger(__name__)


class ExtensionHandoffDAO:
    """DAO for extension_handoffs table.

    States:
      0 = new
      1 = linked
      2 = twitch_connected
    """

    def __init__(self, db: Optional[Db] = None):
        self.db = db if db else Db(DBConfig())

    async def create(
        self, channel_id: str, twitch_login: Optional[str] = None, ttl_minutes: int = 10
    ) -> str:
        """Get or create a new handoff for the channel.

        Idempotent: if an active (status=0) handoff already exists for the channel,
        it will be returned. If it is expired, we renew its expires_at and return it.

        Returns handoff_id (uuid string).
        """
        # Try to find an existing active handoff for this channel
        row = await self.db.fetch_one(
            "SELECT id, expires_at, twitch_login FROM extension_handoffs\n"
            "WHERE channel_id = %s AND status = 0\n"
            "ORDER BY created_at DESC LIMIT 1",
            (channel_id,),
        )
        now = datetime.datetime.now(datetime.timezone.utc)
        new_expiry = now + datetime.timedelta(minutes=ttl_minutes)

        if row:
            existing_id = row[0]
            existing_expires = row[1]
            existing_login = row[2]
            # If expired, renew. Also update twitch_login if newly provided.
            set_cols = []
            set_vals = []
            if existing_expires is None or existing_expires < now:
                set_cols.append("expires_at")
                set_vals.append(new_expiry)
            if twitch_login and twitch_login != existing_login:
                set_cols.append("twitch_login")
                set_vals.append(twitch_login)
            if set_cols:
                await self.db.update(
                    table="extension_handoffs",
                    set_columns=set_cols,
                    set_values=set_vals,
                    where_clause="id = %s",
                    where_params=(existing_id,),
                )
            return str(existing_id)

        # No existing active record → create a new one
        handoff_id = str(uuid.uuid4())
        await self.db.insert(
            table="extension_handoffs",
            columns=[
                "id",
                "channel_id",
                "status",
                "user_id",
                "twitch_login",
                "last_error",
                "created_at",
                "expires_at",
            ],
            values=(
                handoff_id,
                channel_id,
                0,
                None,
                twitch_login,
                None,
                now,
                new_expiry,
            ),
        )
        return str(handoff_id)

    async def get(self, handoff_id: str) -> Optional[tuple[Any, ...]]:
        row = await self.db.fetch_one(
            "SELECT id, channel_id, status, user_id, twitch_login, last_error, created_at, expires_at\n"
            "FROM extension_handoffs WHERE id = %s",
            (handoff_id,),
        )
        return row

    async def mark_linked(self, handoff_id: str, user_id: int) -> None:
        await self.db.update(
            table="extension_handoffs",
            set_columns=["status", "user_id"],
            set_values=[1, user_id],
            where_clause="id = %s",
            where_params=(handoff_id,),
        )

    async def mark_twitch_connected(self, handoff_id: str) -> None:
        await self.db.update(
            table="extension_handoffs",
            set_columns=["status"],
            set_values=[2],
            where_clause="id = %s",
            where_params=(handoff_id,),
        )

    async def set_error(self, handoff_id: str, message: str) -> None:
        await self.db.update(
            table="extension_handoffs",
            set_columns=["last_error"],
            set_values=[message[:1000]],
            where_clause="id = %s",
            where_params=(handoff_id,),
        )

    async def get_last_linked(self, channel_id: str) -> Optional[str]:
        """Return the most recent linked or completed handoff id for a channel.

        Linked states are status in (1, 2): 1=linked, 2=twitch_connected.
        """
        row = await self.db.fetch_one(
            "SELECT id FROM extension_handoffs\n"
            "WHERE channel_id = %s AND status IN (1, 2)\n"
            "ORDER BY created_at DESC LIMIT 1",
            (channel_id,),
        )
        if not row:
            return None
        # fetch_one returns a row tuple where id is at index 0
        return str(row[0])



"""
Models for stream keys.

Stream keys are used for direct stream ingestion and are encrypted at rest.
"""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, ConfigDict


class StreamKeyBase(BaseModel):
    """Base model for stream key data."""

    user_id: int
    name: Optional[str] = None
    description: Optional[str] = None
    is_active: bool = True
    consumer_active: bool = False
    encrypted_key: str


class StreamKeyCreate(StreamKeyBase):
    """Model for creating a new stream key."""

    id: int  # Database ID


class StreamKeyUpdate(BaseModel):
    """Model for updating stream key data."""

    name: Optional[str] = None
    description: Optional[str] = None
    is_active: Optional[bool] = None


class StreamKey(StreamKeyBase):
    """Complete stream key model including database ID and timestamps."""

    id: int  # Database ID
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(from_attributes=True)



from datetime import datetime
from typing import List, Tuple
from typing import Optional

from common.database import DBConfig, Db
from common.config.config import get_pro_model_settings
from common.logging import get_logger, span
from common.models.clip import (
    ClipUpdate,
    Clip,
    ClipCreate,
    ClipAnalytics,
    ClipAnalyticsCreate,
    ClipAnalyticsUpdate,
    EditedClip,
    EditedClipCreate,
    EditedClipUpdate,
    ClipLimitsStatus,
)

logger = get_logger(__name__)


SHOULD_STOP_GENERATING_CLIPS = """
    WITH ctx AS (
    SELECT u.id AS user_id, u.tier
    FROM streams s
    JOIN users   u ON u.id = s.user_id
    WHERE s.id = %s
    ),
    agg AS (
    SELECT
        ctx.user_id,
        ctx.tier,
        COUNT(*) FILTER (
        WHERE c.deleted_at IS NULL
            AND c.created_at::date = CURRENT_DATE
        ) AS clips_today,
        COUNT(*) FILTER (
        WHERE c.deleted_at IS NULL
        ) AS clips_this_month
    FROM ctx
    LEFT JOIN clips c ON c.user_id = ctx.user_id
    GROUP BY ctx.user_id, ctx.tier
    )
    SELECT
    (tier = %s AND (clips_today >= %s OR clips_this_month >= %s)) AS should_stop
    FROM agg;
"""


CLIP_LIMITS_STATUS_BY_USER_ID = """
    SELECT
        COUNT(*) FILTER (
            WHERE deleted_at IS NULL
              AND created_at >= NOW() - INTERVAL '1 day'
        ) AS clips_last_day,
        COUNT(*) FILTER (
            WHERE deleted_at IS NULL
              AND created_at >= NOW() - INTERVAL '30 days'
        ) AS clips_last_month
    FROM clips
    WHERE user_id = %s
"""


CLIP_CORE_SELECT_COLUMNS = """
    c.id,
    c.url,
    c.kind,
    c.thumbnail_url,
    c.watermark_url,
    c.caption,
    c.start_time,
    c.created_at,
    c.updated_at,
    c.deleted_at,
    c.favorited,
    c.views,
    c.stream_id,
    c.owner_username,
    c.user_id,
    c.source_type,
    c.upload_session_id,
    s.platform
"""

CLIP_ANALYTICS_SELECT_COLUMNS = """
    ca.id,
    ca.views,
    ca.heart_reactions,
    ca.smile_reactions,
    ca.laugh_reactions,
    ca.sign_reactions,
    ca.surprised_reactions,
    ca.created_at,
    ca.updated_at,
    ca.deleted_at
"""


class ClipDAO:
    """Data Access Object for clips in PostgreSQL."""

    def __init__(self, db: Optional[Db] = None):
        """Initialize with an existing DB connection or create a new one."""
        self.db = db if db else Db(DBConfig())
        logger.debug("ClipDAO initialized", extra={"component": "ClipDAO"})

    async def create_clip(self, clip: ClipCreate) -> Clip:
        """
        Create a new clip in the database.

        Args:
            clip: The clip data to create

        Returns:
            0 on success, 1 on failure
        """
        clip_attributes = {
            "kind": clip.kind,
            "stream_id": clip.stream_id,
            "owner_username": clip.owner_username,
        }
        with span(logger, "create_clip", clip_attributes):
            columns = [
                "url",
                "kind",
                "stream_id",
                "thumbnail_url",
                "watermark_url",
                "caption",
                "start_time",
                "owner_username",
                "user_id",
                "prompt_version",
                "dimension_scores",
                "combined_score",
                "game_config_name",
                "source_type",
                "upload_session_id",
            ]
            # Prepare a RETURNING clause to get the new clip id.
            # Convert dimension_scores dict to JSON string for JSONB column
            import json

            dimension_scores_json = (
                json.dumps(clip.dimension_scores) if clip.dimension_scores else None
            )

            values = (
                clip.url,
                clip.kind,
                clip.stream_id,
                clip.thumbnail_url,
                clip.watermark_url,
                clip.caption,
                clip.start_time,
                clip.owner_username,
                clip.user_id,
                clip.prompt_version,
                dimension_scores_json,
                clip.combined_score,
                clip.game_config_name,
                clip.source_type,
                str(clip.upload_session_id) if clip.upload_session_id else None,
            )

            try:
                res = await self.db.insert("clips", columns, values)
                if not res:
                    logger.error(
                        f"Failed to create clip of kind '{clip.kind}' for stream {clip.stream_id}"
                    )
                    raise Exception("Failed to create clip")

                # Create clip object with ID
                clip_obj = Clip(
                    id=res,
                    url=clip.url,
                    kind=clip.kind,
                    thumbnail_url=clip.thumbnail_url,
                    watermark_url=clip.watermark_url,
                    caption=clip.caption,
                    start_time=clip.start_time,
                    created_at=datetime.now(),
                    updated_at=datetime.now(),
                    deleted_at=None,
                    favorited=False,
                    views=0,
                    stream_id=clip.stream_id,
                    owner_username=clip.owner_username,
                    user_id=clip.user_id,
                    source_type=clip.source_type,
                    upload_session_id=clip.upload_session_id,
                    prompt_version=clip.prompt_version,
                    dimension_scores=clip.dimension_scores,
                    combined_score=clip.combined_score,
                    game_config_name=clip.game_config_name,
                )

                # Log success with complete details
                logger.info(
                    "Clip created successfully",
                    extra={
                        "clip_id": clip_obj.id,
                        "kind": clip_obj.kind,
                        "stream_id": clip_obj.stream_id,
                        "url": clip_obj.url,
                        "thumbnail_url": clip_obj.thumbnail_url,
                        "owner_username": clip_obj.owner_username,
                        "user_id": clip_obj.user_id,
                    },
                )

                return clip_obj
            except Exception as e:
                logger.error(f"Error creating clip: {e}")
                raise

    async def get_edited_clips_by_user_id(
        self, user_id: int, limit: int = 10, offset: int = 0
    ) -> List[Clip]:
        """Get edited clips by user ID with pagination."""
        with span(
            logger,
            "get_edited_clips_by_user_id",
            {"user_id": user_id, "limit": limit, "offset": offset},
        ):
            query = """
                    SELECT ec.id,
                           ec.url,
                           c.kind,
                           ec.thumbnail_url,
                           c.watermark_url,
                           ec.name as caption,
                           c.start_time,
                           ec.created_at,
                           c.updated_at,
                           c.deleted_at,
                           c.favorited,
                           c.views,
                           c.stream_id,
                           c.owner_username,
                           c.user_id,
                           c.source_type,
                           c.upload_session_id,
                           s.platform
                    FROM clips c
                    INNER JOIN edited_clips ec ON c.id = ec.original_clip_id
                    INNER JOIN streams s on s.id = c.stream_id
                    WHERE c.user_id = %s
                      AND c.deleted_at IS NULL AND ec.deleted_at IS NULL
                    ORDER BY ec.created_at DESC
                        LIMIT %s
                    OFFSET %s \
                    """
            result = await self.db.fetch_all(query, (user_id, limit, offset))
            if not result:
                logger.debug(
                    f"No edited clips found for user ID {user_id}",
                    extra={"user_id": user_id},
                )
                return []
            clips = [self._row_to_clip(row) for row in result]
            logger.debug(
                f"Found {len(clips)} edited clips for user ID {user_id}",
                extra={"user_id": user_id, "clip_count": len(clips)},
            )
            return clips

    async def get_clips_by_user_id(
        self, user_id: int, limit: int = 10, offset: int = 0
    ) -> List[Clip]:
        """Get clips by user ID with pagination."""
        with span(
            logger,
            "get_clips_by_user_id",
            {"user_id": user_id, "limit": limit, "offset": offset},
            log_entry=False,
            log_exit=False,
        ):
            query = f"""
                    SELECT {CLIP_CORE_SELECT_COLUMNS},
                           {CLIP_ANALYTICS_SELECT_COLUMNS}
                    FROM clips c
                    INNER JOIN streams s on s.id = c.stream_id
                    LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                    WHERE c.user_id = %s
                      AND c.deleted_at IS NULL
                    ORDER BY c.created_at DESC
                        LIMIT %s
                    OFFSET %s \
                    """
            result = await self.db.fetch_all(query, (user_id, limit, offset))
            if not result:
                logger.debug(
                    f"No clips found for user ID {user_id}", extra={"user_id": user_id}
                )
                return []
            clips = [self._row_to_clip(row) for row in result]
            logger.debug(
                f"Found {len(clips)} clips for user ID {user_id}",
                extra={"user_id": user_id, "clip_count": len(clips)},
            )
            return clips

    async def get_clip_limits_status(self, user_id: int) -> ClipLimitsStatus:
        """Return rolling clip counts (last 24h / last 30d) plus configured limits."""
        with span(
            logger,
            "get_clip_limits_status",
            {"user_id": user_id},
            log_entry=False,
            log_exit=False,
        ):
            settings = get_pro_model_settings()

            row = await self.db.fetch_one(CLIP_LIMITS_STATUS_BY_USER_ID, (user_id,))
            clips_last_day = int((row[0] if row else 0) or 0)
            clips_last_month = int((row[1] if row else 0) or 0)

            return ClipLimitsStatus(
                clips_last_day=clips_last_day,
                clips_last_month=clips_last_month,
                max_clips_per_day=int(settings.max_clips_per_day),
                max_clips_per_month=int(settings.max_clips_per_month),
            )

    async def count_user_clips_this_month(self, user_id: int) -> int:
        """Count non-deleted clips a user has created in the current calendar month."""

        with span(
            logger,
            "count_user_clips_this_month",
            {"user_id": user_id},
            log_entry=False,
            log_exit=False,
        ):
            query = """
                SELECT COUNT(*)
                FROM clips
                WHERE user_id = %s
                  AND deleted_at IS NULL
                  AND date_trunc('month', created_at AT TIME ZONE 'utc') = date_trunc('month', NOW() AT TIME ZONE 'utc')
            """

            result = await self.db.fetch_one(query, (user_id,))
            if not result:
                logger.debug(
                    "No clips found for user in current month",
                    extra={"user_id": user_id},
                )
                return 0

            count = int(result[0])
            logger.debug(
                "Counted user clips for current month",
                extra={"user_id": user_id, "clip_count": count},
            )
            return count

    async def get_clips_by_user_id_paginated(
        self, user_id: int, cursor: Optional[str] = None, limit: int = 20
    ) -> Tuple[List[Clip], Optional[str], bool]:
        """
        Get clips by user ID with cursor-based pagination.

        Args:
            user_id: User ID to fetch clips for
            cursor: Base64-encoded cursor from previous page (format: "id_timestamp")
            limit: Maximum number of clips to return (default 20)

        Returns:
            Tuple of (clips, next_cursor, has_more)
        """
        with span(
            logger,
            "get_clips_by_user_id_paginated",
            {"user_id": user_id, "has_cursor": cursor is not None, "limit": limit},
            log_entry=False,
            log_exit=False,
        ):
            # Decode cursor if present
            if cursor:
                from base64 import b64decode

                try:
                    decoded = b64decode(cursor).decode("utf-8")
                    cursor_id, cursor_timestamp = decoded.split("_")
                    cursor_id = int(cursor_id)
                    cursor_timestamp = int(cursor_timestamp)

                    query = f"""
                        SELECT {CLIP_CORE_SELECT_COLUMNS},
                               {CLIP_ANALYTICS_SELECT_COLUMNS}
                        FROM clips c
                        INNER JOIN streams s on s.id = c.stream_id
                        LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                        WHERE c.user_id = %s
                          AND c.deleted_at IS NULL
                          AND (EXTRACT(EPOCH FROM c.created_at), c.id) < (%s, %s)
                        ORDER BY c.created_at DESC, c.id DESC
                        LIMIT %s
                    """
                    result = await self.db.fetch_all(
                        query, (user_id, cursor_timestamp, cursor_id, limit + 1)
                    )
                except Exception as e:
                    logger.error(
                        f"Invalid cursor format: {e}",
                        extra={"user_id": user_id, "cursor": cursor},
                    )
                    # Fall back to first page
                    cursor = None

            if not cursor:
                # First page
                query = f"""
                    SELECT {CLIP_CORE_SELECT_COLUMNS},
                           {CLIP_ANALYTICS_SELECT_COLUMNS}
                    FROM clips c
                    INNER JOIN streams s on s.id = c.stream_id
                    LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                    WHERE c.user_id = %s
                      AND c.deleted_at IS NULL
                    ORDER BY c.created_at DESC, c.id DESC
                    LIMIT %s
                """
                result = await self.db.fetch_all(query, (user_id, limit + 1))

            if not result:
                logger.debug(
                    f"No clips found for user ID {user_id}", extra={"user_id": user_id}
                )
                return [], None, False

            # Check if there are more results
            has_more = len(result) > limit
            items = result[:limit]

            # Generate next cursor from last item
            next_cursor = None
            if has_more and items:
                from base64 import b64encode

                last_item = items[-1]
                last_id = last_item[0]
                last_created_at = last_item[7]
                cursor_str = f"{last_id}_{int(last_created_at.timestamp())}"
                next_cursor = b64encode(cursor_str.encode("utf-8")).decode("utf-8")

            clips = [self._row_to_clip(row) for row in items]
            logger.debug(
                f"Found {len(clips)} clips for user ID {user_id}",
                extra={
                    "user_id": user_id,
                    "clip_count": len(clips),
                    "has_more": has_more,
                },
            )
            return clips, next_cursor, has_more

    async def get_edited_clips_by_user_id_paginated(
        self, user_id: int, cursor: Optional[str] = None, limit: int = 20
    ) -> Tuple[List[Clip], Optional[str], bool]:
        """
        Get edited clips by user ID with cursor-based pagination.

        Args:
            user_id: User ID to fetch edited clips for
            cursor: Base64-encoded cursor from previous page (format: "id_timestamp")
            limit: Maximum number of clips to return (default 20)

        Returns:
            Tuple of (clips, next_cursor, has_more)
        """
        with span(
            logger,
            "get_edited_clips_by_user_id_paginated",
            {"user_id": user_id, "has_cursor": cursor is not None, "limit": limit},
            log_entry=False,
            log_exit=False,
        ):
            # Decode cursor if present
            if cursor:
                from base64 import b64decode

                try:
                    decoded = b64decode(cursor).decode("utf-8")
                    cursor_id, cursor_timestamp = decoded.split("_")
                    cursor_id = int(cursor_id)
                    cursor_timestamp = int(cursor_timestamp)

                    query = """
                        SELECT ec.id,
                               ec.url,
                               c.kind,
                               ec.thumbnail_url,
                               c.watermark_url,
                               ec.name as caption,
                               c.start_time,
                               ec.created_at,
                               c.updated_at,
                               c.deleted_at,
                               c.favorited,
                               c.views,
                               c.stream_id,
                               c.owner_username,
                               c.user_id,
                               c.source_type,
                               c.upload_session_id,
                               s.platform
                        FROM clips c
                        INNER JOIN edited_clips ec ON c.id = ec.original_clip_id
                        INNER JOIN streams s on s.id = c.stream_id
                        WHERE c.user_id = %s
                          AND c.deleted_at IS NULL AND ec.deleted_at IS NULL
                          AND (EXTRACT(EPOCH FROM ec.created_at), ec.id) < (%s, %s)
                        ORDER BY ec.created_at DESC, ec.id DESC
                        LIMIT %s
                    """
                    result = await self.db.fetch_all(
                        query, (user_id, cursor_timestamp, cursor_id, limit + 1)
                    )
                except Exception as e:
                    logger.error(
                        f"Invalid cursor format: {e}",
                        extra={"user_id": user_id, "cursor": cursor},
                    )
                    # Fall back to first page
                    cursor = None

            if not cursor:
                # First page
                query = """
                    SELECT ec.id,
                           ec.url,
                           c.kind,
                           ec.thumbnail_url,
                           c.watermark_url,
                           ec.name as caption,
                           c.start_time,
                           ec.created_at,
                           c.updated_at,
                           c.deleted_at,
                           c.favorited,
                           c.views,
                           c.stream_id,
                           c.owner_username,
                           c.user_id,
                           c.source_type,
                           c.upload_session_id,
                           s.platform
                    FROM clips c
                    INNER JOIN edited_clips ec ON c.id = ec.original_clip_id
                    INNER JOIN streams s on s.id = c.stream_id
                    WHERE c.user_id = %s
                      AND c.deleted_at IS NULL AND ec.deleted_at IS NULL
                    ORDER BY ec.created_at DESC, ec.id DESC
                    LIMIT %s
                """
                result = await self.db.fetch_all(query, (user_id, limit + 1))

            if not result:
                logger.debug(
                    f"No edited clips found for user ID {user_id}",
                    extra={"user_id": user_id},
                )
                return [], None, False

            # Check if there are more results
            has_more = len(result) > limit
            items = result[:limit]

            # Generate next cursor from last item
            next_cursor = None
            if has_more and items:
                from base64 import b64encode

                last_item = items[-1]
                last_id = last_item[0]
                last_created_at = last_item[7]
                cursor_str = f"{last_id}_{int(last_created_at.timestamp())}"
                next_cursor = b64encode(cursor_str.encode("utf-8")).decode("utf-8")

            clips = [self._row_to_clip(row) for row in items]
            logger.debug(
                f"Found {len(clips)} edited clips for user ID {user_id}",
                extra={
                    "user_id": user_id,
                    "clip_count": len(clips),
                    "has_more": has_more,
                },
            )
            return clips, next_cursor, has_more

    async def get_clips_by_username(
        self, username: str, limit: int = 10, offset: int = 1
    ) -> List[Clip]:
        """Get clips by owner username with pagination."""
        with span(
            logger,
            "get_clips_by_username",
            {"username": username, "limit": limit, "offset": offset},
        ):
            query = f"""
                    SELECT {CLIP_CORE_SELECT_COLUMNS},
                           {CLIP_ANALYTICS_SELECT_COLUMNS}
                    FROM clips c
                    INNER JOIN streams s on s.id = c.stream_id
                    LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                    WHERE c.owner_username = %s
                      AND c.deleted_at IS NULL
                    ORDER BY c.created_at DESC
                        LIMIT %s
                    OFFSET %s \
                    """
            result = await self.db.fetch_all(query, (username, limit, offset))
            if not result:
                logger.debug(
                    f"No clips found for username {username}",
                    extra={"username": username},
                )
                return []
            clips = [self._row_to_clip(row) for row in result]
            logger.debug(
                f"Found {len(clips)} clips for username {username}",
                extra={"username": username, "clip_count": len(clips)},
            )
            return clips

    async def get_clips_by_stream_ids(self, stream_ids: List[str]) -> List[Clip]:
        """Get all clips for the specified stream IDs sorted by start time."""
        with span(
            logger,
            "get_clips_by_stream_ids",
            {"stream_ids_count": len(stream_ids) if stream_ids else 0},
        ):
            if not stream_ids:
                logger.debug("No stream IDs provided, returning empty list")
                return []

            rows = await self.db.fetch_all(
                f"""
                SELECT {CLIP_CORE_SELECT_COLUMNS},
                       {CLIP_ANALYTICS_SELECT_COLUMNS}
                FROM clips c
                INNER JOIN streams s on s.id = c.stream_id
                LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                WHERE c.stream_id = ANY (%s)
                  AND c.deleted_at IS NULL
                ORDER BY c.start_time ASC
                """,
                (stream_ids,),
            )

            if not rows:
                logger.debug(
                    "No clips found for provided stream IDs",
                    extra={"stream_ids_count": len(stream_ids)},
                )
                return []

            clips = [self._row_to_clip(row) for row in rows]
            logger.debug(
                f"Found {len(clips)} clips for {len(stream_ids)} stream IDs",
                extra={"stream_ids_count": len(stream_ids), "clip_count": len(clips)},
            )
            return clips

    async def should_stop_generating_clips(self, stream_id: str) -> bool:
        """Validates for free tier users should generating clips or not"""
        """True: stop generating, False: good to go"""
        with span(logger, "should_stop_generating_clips"):
            settings = get_pro_model_settings()
            if not settings.pro_feature_enable:
                return False
            try:
                row = await self.db.fetch_one(
                    SHOULD_STOP_GENERATING_CLIPS,
                    (
                        stream_id,
                        settings.free_tier,
                        settings.max_clips_per_day,
                        settings.max_clips_per_month,
                    ),
                )
                if not row:
                    logger.error(
                        f"Error fetching database info for generate clip validation on stream {stream_id}"
                    )
                    return False
                else:
                    return bool(row[0])
            except Exception as e:
                logger.error(
                    f"Error validating for free tier users if should generating clips on stream {stream_id}, error message: {e}"
                )
                raise

    async def get_clip_by_id(self, clip_id: int) -> Optional[Clip]:
        """Get a clip by its ID."""
        with span(logger, "get_clip_by_id", {"clip_id": clip_id}):
            row = await self.db.fetch_one(
                f"""
                                          SELECT {CLIP_CORE_SELECT_COLUMNS},
                                                 {CLIP_ANALYTICS_SELECT_COLUMNS}
                                          FROM clips c
                                          INNER JOIN streams s on s.id = c.stream_id
                                          LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                                          WHERE c.id = %s
                                            AND c.deleted_at IS NULL
                                          """,
                (clip_id,),
            )

            if not row:
                logger.debug(
                    f"Clip with ID {clip_id} not found", extra={"clip_id": clip_id}
                )
                return None

            clip = self._row_to_clip(row)
            logger.debug(
                f"Found clip with ID {clip_id}",
                extra={"clip_id": clip_id, "clip_kind": clip.kind},
            )
            return clip

    async def get_clips_by_stream(
        self, stream_id: str, limit: int = 50, offset: int = 0
    ):
        """Get all clips for a specific stream."""
        with span(
            logger,
            "get_clips_by_stream",
            {"stream_id": stream_id, "limit": limit, "offset": offset},
        ):
            rows = await self.db.fetch_all(
                f"""
                                           SELECT {CLIP_CORE_SELECT_COLUMNS},
                                                  {CLIP_ANALYTICS_SELECT_COLUMNS}
                                           FROM clips c
                                           INNER JOIN streams s on s.id = c.stream_id
                                           LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                                           WHERE c.stream_id = %s
                                             AND c.deleted_at IS NULL
                                           ORDER BY c.start_time ASC
                                               LIMIT %s
                                           OFFSET %s
                                           """,
                (stream_id, limit, offset),
            )

            if not rows:
                logger.debug(
                    f"No clips found for stream ID {stream_id}",
                    extra={"stream_id": stream_id},
                )
                return []

            clips = [self._row_to_clip(row) for row in rows]
            logger.debug(
                f"Found {len(clips)} clips for stream ID {stream_id}",
                extra={"stream_id": stream_id, "clip_count": len(clips)},
            )
            return clips

    async def get_clips_by_kind(
        self, kind: str, limit: int = 50, offset: int = 0
    ) -> List[Clip]:
        """Get clips by their type/kind."""
        with span(
            logger,
            "get_clips_by_kind",
            {"kind": kind, "limit": limit, "offset": offset},
        ):
            rows = await self.db.fetch_all(
                f"""
                                           SELECT {CLIP_CORE_SELECT_COLUMNS},
                                                  {CLIP_ANALYTICS_SELECT_COLUMNS}
                                           FROM clips c
                                           INNER JOIN streams s on s.id = c.stream_id
                                           LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                                           WHERE c.kind = %s
                                             AND c.deleted_at IS NULL
                                           ORDER BY c.created_at DESC
                                               LIMIT %s
                                           OFFSET %s
                                           """,
                (kind, limit, offset),
            )
            if not rows:
                logger.debug(f"No clips found for kind '{kind}'", extra={"kind": kind})
                return []

            clips = [self._row_to_clip(row) for row in rows]
            logger.debug(
                f"Found {len(clips)} clips for kind '{kind}'",
                extra={"kind": kind, "clip_count": len(clips)},
            )
            return clips

    async def get_clips_after_time(self, timestamp: int) -> List[Clip]:
        """Get clips created after a specific timestamp."""
        with span(logger, "get_clips_after_time", {"timestamp": timestamp}):
            rows = await self.db.fetch_all(
                f"""
                                           SELECT {CLIP_CORE_SELECT_COLUMNS},
                                                  {CLIP_ANALYTICS_SELECT_COLUMNS}
                                           FROM clips c
                                           INNER JOIN streams s on s.id = c.stream_id
                                           LEFT JOIN clip_analytics ca on ca.clip_id = c.id AND ca.deleted_at IS NULL
                                           WHERE c.start_time >= %s
                                             AND c.deleted_at IS NULL
                                           ORDER BY c.start_time DESC
                                           """,
                (timestamp,),
            )
            if not rows:
                logger.debug(
                    f"No clips found after timestamp {timestamp}",
                    extra={"timestamp": timestamp},
                )
                return []

            clips = [self._row_to_clip(row) for row in rows]
            logger.debug(
                f"Found {len(clips)} clips after timestamp {timestamp}",
                extra={"timestamp": timestamp, "clip_count": len(clips)},
            )
            return clips

    async def count_clips_by_stream_and_kind(self, stream_id: int, kind: str) -> int:
        """Count clips by internal stream ID and kind."""
        with span(
            logger,
            "count_clips_by_stream_and_kind",
            {"stream_id": stream_id, "kind": kind},
        ):
            row = await self.db.fetch_one(
                """
                                          SELECT COUNT(*)
                                          FROM clips
                                          WHERE stream_id = %s
                                            AND LOWER(kind) LIKE %s
                                            AND deleted_at IS NULL
                                          """,
                (stream_id, f"%{kind.lower()}%"),
            )
            count = row[0] if row else 0
            logger.debug(
                f"Counted {count} clips for stream ID {stream_id} with kind '{kind}'",
                extra={"stream_id": stream_id, "kind": kind, "count": count},
            )
            return count

    async def count_clips_by_stream(self, stream_id: int) -> int:
        """Count clips by internal stream ID."""
        with span(logger, "count_clips_by_stream", {"stream_id": stream_id}):
            row = await self.db.fetch_one(
                """
                                          SELECT COUNT(*)
                                          FROM clips
                                          WHERE stream_id = %s
                                            AND deleted_at IS NULL
                                          """,
                (stream_id,),
            )
            count = row[0] if row else 0
            logger.debug(
                f"Counted {count} clips for stream ID {stream_id}",
                extra={"stream_id": stream_id, "count": count},
            )
            return count

    async def has_one_clip(self, stream_id: int) -> bool:
        """Return True if the given internal stream ID has exactly one non-deleted clip."""
        with span(logger, "has_one_clip", {"stream_id": stream_id}):
            try:
                count = await self.count_clips_by_stream(stream_id)
                logger.debug(
                    "Checked clip count for stream",
                    extra={"stream_id": stream_id, "clip_count": count},
                )
                return count == 1
            except Exception as e:
                logger.error(
                    "Error checking if stream has a single clip",
                    extra={"stream_id": stream_id, "error": str(e)},
                )
                raise

    async def count_clip_views_by_stream(self, stream_id: int) -> int:
        """Count clip views by internal stream ID."""
        with span(logger, "count_clip_views_by_stream", {"stream_id": stream_id}):
            row = await self.db.fetch_one(
                """
                                          SELECT SUM(views)
                                          FROM clips
                                          WHERE stream_id = %s
                                            AND deleted_at IS NULL
                                          """,
                (stream_id,),
            )
            count = row[0] if row and row[0] is not None else 0
            logger.debug(
                f"Counted {count} total views for stream ID {stream_id}",
                extra={"stream_id": stream_id, "view_count": count},
            )
            return count

    async def update_clip(self, clip_id: int, update: ClipUpdate) -> None:
        """
        Update clip details.

        Args:
            clip_id: The ID of the clip to update
            update: The update data
        """
        update_dict = update.dict(exclude_none=True)
        if not update_dict:
            logger.debug(f"No updates provided for clip ID {clip_id}, skipping")
            return

        with span(
            logger,
            "update_clip",
            {"clip_id": clip_id, "fields": list(update_dict.keys())},
        ):
            set_clauses = ["updated_at = CURRENT_TIMESTAMP"]
            values = []

            for key, value in update_dict.items():
                set_clauses.append(key)
                values.append(value)

            try:
                await self.db.update(
                    table="clips",
                    set_columns=set_clauses,
                    set_values=values,
                    where_clause="id = %s AND deleted_at IS NULL",
                )
                values.append(clip_id)
                logger.info(
                    f"Successfully updated clip ID {clip_id} with fields: {list(update_dict.keys())}"
                )
            except Exception as e:
                logger.error(f"Error updating clip ID {clip_id}: {e}")
                raise

    async def soft_delete_clip(self, clip_id: int) -> None:
        """
        Soft delete a clip by setting deleted_at.

        Args:
            clip_id: The ID of the clip to soft delete
        """
        with span(logger, "soft_delete_clip", {"clip_id": clip_id}):
            try:
                await self.db.execute_commit(
                    """
                                             UPDATE clips
                                             SET deleted_at = CURRENT_TIMESTAMP
                                             WHERE id = %s
                                               AND deleted_at IS NULL
                                             """,
                    (clip_id,),
                )
                logger.info(f"Successfully soft deleted clip ID {clip_id}")
            except Exception as e:
                logger.error(f"Error soft deleting clip ID {clip_id}: {e}")
                raise

    async def hard_delete_clip(self, clip_id: int) -> None:
        """
        Hard delete a clip from the database.

        Args:
            clip_id: The ID of the clip to permanently delete
        """
        with span(logger, "hard_delete_clip", {"clip_id": clip_id}):
            logger.warning(f"Permanently deleting clip ID {clip_id} from database")
            try:
                await self.db.execute_commit(
                    """
                                             DELETE
                                             FROM clips
                                             WHERE id = %s
                                             """,
                    (clip_id,),
                )
                logger.info(f"Successfully hard deleted clip ID {clip_id}")
            except Exception as e:
                logger.error(f"Error hard deleting clip ID {clip_id}: {e}")
                raise

    async def increment_view_count(self, clip_id: int) -> None:
        """
        Increment the view count of a clip by 1.

        Args:
            clip_id: The ID of the clip to increment views for
        """
        with span(logger, "increment_view_count", {"clip_id": clip_id}):
            try:
                await self.db.execute_commit(
                    """
                                             UPDATE clips
                                             SET views      = views + 1,
                                                 updated_at = CURRENT_TIMESTAMP
                                             WHERE id = %s
                                               AND deleted_at IS NULL
                                             """,
                    (clip_id,),
                )
            except Exception as e:
                logger.error(
                    f"Error incrementing view count for clip ID {clip_id}: {e}"
                )
                raise

    def _row_to_clip(self, row) -> Clip:
        """
        Map columns as:
          0: id
          1: url
          2: kind
          3: thumbnail_url
          4: watermark_url
          5: caption
          6: start_time (int)
          7: created_at (datetime)
          8: updated_at (datetime)
          9: deleted_at (datetime or None)
          10: favorited
          11: views
          12: stream_id (int)
          13: owner_username (string)
          14: user_id (int)
          15: source_type
          16: upload_session_id (UUID or None)
          17: platform (text) from streams table
          18 - 27: analytics
        """
        caption = str(row[5]) if row[5] is not None else ""
        start_time = int(row[6])
        created_at = row[7]
        updated_at = row[8] if isinstance(row[8], datetime) else datetime.now()
        deleted_at = row[9] if isinstance(row[9], datetime) else None

        # Safely handle presence or absence of analytics columns.
        has_analytics = len(row) > 18 and row[18] is not None

        return Clip(
            id=row[0],
            url=row[1],
            kind=row[2],
            thumbnail_url=row[3],
            watermark_url=row[4],
            caption=caption,
            start_time=start_time,
            created_at=created_at,
            updated_at=updated_at,
            deleted_at=deleted_at,
            favorited=row[10] if row[10] is not None else False,
            views=int(row[11]),
            stream_id=int(row[12]),
            owner_username=str(row[13]) if row[13] is not None else "",
            user_id=int(row[14]) if row[14] is not None else -1,
            source_type=row[15] if row[15] is not None else "stream",
            upload_session_id=row[16],
            platform=row[17] if row[17] is not None else None,
            analytics=(
                ClipAnalytics(
                    id=row[18],
                    clip_id=row[0],
                    views=row[19],
                    heart_reactions=row[20],
                    smile_reactions=row[21],
                    laugh_reactions=row[22],
                    sign_reactions=row[23],
                    surprised_reactions=row[24],
                    created_at=row[25],
                    updated_at=row[26],
                    deleted_at=row[27],
                )
                if has_analytics
                else None
            ),
        )

    # Edited Clips Methods

    async def create_edited_clip(self, clip: EditedClipCreate) -> EditedClip:
        """
        Create a new edited clip in the database.

        Args:
            clip: The edited clip data to create

        Returns:
            The created edited clip with its ID and timestamps

        Raises:
            Exception: If the clip creation fails
        """
        with span(
            logger, "create_edited_clip", {"original_clip_id": clip.original_clip_id}
        ):
            columns = ["original_clip_id", "url", "name", "metadata", "thumbnail_url"]
            values = (
                clip.original_clip_id,
                clip.url,
                clip.name,
                clip.metadata,
                clip.thumbnail_url,
            )

            try:
                clip_id = await self.db.insert("edited_clips", columns, values)
                if not clip_id:
                    logger.error(
                        f"Failed to create edited clip for original clip {clip.original_clip_id}"
                    )
                    raise Exception("Failed to create edited clip")

                logger.info(
                    f"Successfully created edited clip for original clip {clip.original_clip_id}"
                )

                # Fetch the created clip to return with all fields
                created_clip = await self.get_edited_clip_by_id(clip_id)
                if not created_clip:
                    logger.error(f"Failed to fetch created edited clip ID {clip_id}")
                    raise Exception("Failed to fetch created edited clip")
                return created_clip
            except Exception as e:
                logger.error(f"Error creating edited clip: {e}")
                raise

    async def get_edited_clip_by_id(self, clip_id: int) -> Optional[EditedClip]:
        """
        Get an edited clip by its ID.

        Args:
            clip_id: The ID of the edited clip to retrieve

        Returns:
            The edited clip if found, None otherwise
        """
        with span(logger, "get_edited_clip_by_id", {"clip_id": clip_id}):
            row = await self.db.fetch_one(
                """
                SELECT id, original_clip_id, url, name, metadata, created_at, updated_at, deleted_at, thumbnail_url
                FROM edited_clips
                WHERE id = %s AND deleted_at IS NULL
            """,
                (clip_id,),
            )

            if not row:
                logger.debug(f"Edited clip with ID {clip_id} not found")
                return None

            edited_clip = self._row_to_edited_clip(row)
            logger.debug(f"Found edited clip with ID {clip_id}")
            return edited_clip

    async def get_edited_clips_by_original_clip_id(
        self, original_clip_id: int
    ) -> List[EditedClip]:
        """
        Get all edited versions of an original clip.

        Args:
            original_clip_id: The ID of the original clip

        Returns:
            A list of edited clips for the original clip
        """
        with span(
            logger,
            "get_edited_clips_by_original_clip_id",
            {"original_clip_id": original_clip_id},
        ):
            rows = await self.db.fetch_all(
                """
                SELECT id, original_clip_id, url, name, metadata, created_at, updated_at, deleted_at, thumbnail_url
                FROM edited_clips
                WHERE original_clip_id = %s AND deleted_at IS NULL
                ORDER BY created_at DESC
            """,
                (original_clip_id,),
            )

            if not rows:
                logger.debug(
                    f"No edited clips found for original clip ID {original_clip_id}"
                )
                return []

            edited_clips = [self._row_to_edited_clip(row) for row in rows]
            logger.debug(
                f"Found {len(edited_clips)} edited clips for original clip ID {original_clip_id}"
            )
            return edited_clips

    async def update_edited_clip(self, clip_id: int, update: EditedClipUpdate) -> None:
        """
        Update an edited clip.

        Args:
            clip_id: The ID of the edited clip to update
            update: The update data

        Raises:
            Exception: If the update fails
        """
        update_dict = update.model_dump(exclude_none=True)
        if not update_dict:
            logger.debug(f"No updates provided for edited clip ID {clip_id}, skipping")
            return

        with span(
            logger,
            "update_edited_clip",
            {"clip_id": clip_id, "fields": list(update_dict.keys())},
        ):
            set_clauses = []
            values = []

            for key, value in update_dict.items():
                set_clauses.append(key)
                values.append(value)

            try:
                await self.db.update(
                    table="edited_clips",
                    set_columns=set_clauses,
                    set_values=values,
                    where_clause="id = %s AND deleted_at IS NULL",
                    where_params=(clip_id,),
                )
                # values.append(clip_id)
                logger.info(
                    f"Successfully updated edited clip ID {clip_id} with fields: {list(update_dict.keys())}"
                )
            except Exception as e:
                logger.error(f"Error updating edited clip ID {clip_id}: {e}")
                raise

    async def soft_delete_edited_clip(self, clip_id: int) -> None:
        """
        Soft delete an edited clip by setting deleted_at.

        Args:
            clip_id: The ID of the edited clip to soft delete

        Raises:
            Exception: If the deletion fails
        """
        with span(logger, "soft_delete_edited_clip", {"clip_id": clip_id}):
            try:
                await self.db.execute_commit(
                    """
                    UPDATE edited_clips
                    SET deleted_at = CURRENT_TIMESTAMP
                    WHERE id = %s AND deleted_at IS NULL
                """,
                    (clip_id,),
                )
                logger.info(f"Successfully soft deleted edited clip ID {clip_id}")
            except Exception as e:
                logger.error(f"Error soft deleting edited clip ID {clip_id}: {e}")
                raise

    def _row_to_edited_clip(self, row) -> EditedClip:
        """
        Convert a database row to an EditedClip model.

        Args:
            row: The database row

        Returns:
            An EditedClip model
        """
        return EditedClip(
            id=row[0],
            original_clip_id=row[1],
            url=row[2],
            name=row[3],
            metadata=row[4],
            created_at=row[5],
            updated_at=row[6],
            deleted_at=row[7],
            thumbnail_url=row[8],
        )


class ClipAnalyticsDAO:
    """Data Access Object for clip analytics in PostgreSQL."""

    def __init__(self, db: Db):
        # Initialize with an existing DB connection or create a new one.
        self.db = db if db else Db(DBConfig())
        logger.debug(
            "ClipAnalyticsDAO initialized", extra={"component": "ClipAnalyticsDAO"}
        )

    async def create_clip_analytics(
        self, analytics: ClipAnalyticsCreate
    ) -> Optional[ClipAnalytics]:
        """
        Create a new clip analytics record.
        """
        with span(logger, "create_clip_analytics", {"clip_id": analytics.clip_id}):
            columns = [
                "clip_id",
                "views",
                "heart_reactions",
                "smile_reactions",
                "laugh_reactions",
                "sign_reactions",
                "surprised_reactions",
            ]
            values = (
                analytics.clip_id,
                analytics.views,
                analytics.heart_reactions,
                analytics.smile_reactions,
                analytics.laugh_reactions,
                analytics.sign_reactions,
                analytics.surprised_reactions,
            )

            try:
                result = await self.db.insert(
                    table="clip_analytics", columns=columns, values=values
                )
                if result:
                    logger.info(
                        f"Created analytics for clip ID {analytics.clip_id}",
                        extra={
                            "clip_id": analytics.clip_id,
                            "views": analytics.views,
                            "heart_reactions": analytics.heart_reactions,
                        },
                    )
                    # Commented code preserved as it might be needed in the future
                    # return ClipAnalytics(
                    #     id=result[0],
                    #     clip_id=analytics.clip_id,
                    #     views=analytics.views,
                    #     hearts=analytics.hearts,
                    #     created_at=result[1],
                    #     updated_at=result[2],
                    #     deleted_at=None
                    # )
                else:
                    logger.error(
                        f"Failed to create analytics for clip ID {analytics.clip_id}",
                        extra={"clip_id": analytics.clip_id},
                    )
            except Exception as e:
                logger.error(
                    f"Error creating analytics for clip ID {analytics.clip_id}: {e}",
                    extra={"clip_id": analytics.clip_id, "error": str(e)},
                )

            return None

    async def get_clip_analytics_by_clip_id(
        self, clip_id: int
    ) -> Optional[ClipAnalytics]:
        """
        Retrieve the analytics for a specific clip using its ID.
        """
        with span(logger, "get_clip_analytics_by_clip_id", {"clip_id": clip_id}):
            try:
                row = await self.db.fetch_one(
                    """
                                              SELECT *
                                              FROM clip_analytics
                                              WHERE clip_id = %s
                                                AND deleted_at IS NULL
                                              """,
                    (clip_id,),
                )
                if not row:
                    logger.debug(
                        f"No analytics found for clip ID {clip_id}",
                        extra={"clip_id": clip_id},
                    )
                    return None

                analytics = self._row_to_clip_analytics(row)
                logger.debug(
                    f"Retrieved analytics for clip ID {clip_id}",
                    extra={
                        "clip_id": clip_id,
                        "views": analytics.views,
                        "heart_reactions": analytics.heart_reactions,
                    },
                )
                return analytics
            except Exception as e:
                logger.error(
                    f"Error retrieving analytics for clip ID {clip_id}: {e}",
                    extra={"clip_id": clip_id, "error": str(e)},
                )
                return None

    async def update_clip_analytics(
        self, analytics_id: int, update: ClipAnalyticsUpdate
    ) -> None:
        """
        Update clip analytics details.
        """
        with span(logger, "update_clip_analytics", {"analytics_id": analytics_id}):
            update_dict = update.dict(exclude_none=True)
            if not update_dict:
                logger.debug(
                    f"No updates provided for analytics ID {analytics_id}, skipping",
                    extra={"analytics_id": analytics_id},
                )
                return

            set_clauses = ["updated_at = CURRENT_TIMESTAMP"]
            values = []
            for key, value in update_dict.items():
                set_clauses.append(key)
                values.append(value)

            try:
                # Update the clip analytics record.
                await self.db.update(
                    table="clip_analytics",
                    set_columns=set_clauses,
                    set_values=values,
                    where_clause="id = %s AND deleted_at IS NULL",
                )
                values.append(analytics_id)
                logger.info(
                    f"Updated analytics ID {analytics_id} with fields: {list(update_dict.keys())}",
                    extra={
                        "analytics_id": analytics_id,
                        "fields": list(update_dict.keys()),
                    },
                )
            except Exception as e:
                logger.error(
                    f"Error updating analytics ID {analytics_id}: {e}",
                    extra={"analytics_id": analytics_id, "error": str(e)},
                )
                raise

    async def soft_delete_clip_analytics(self, analytics_id: int) -> None:
        """
        Soft delete a clip analytics record by setting deleted_at.
        """
        with span(logger, "soft_delete_clip_analytics", {"analytics_id": analytics_id}):
            try:
                await self.db.execute_commit(
                    """
                                             UPDATE clip_analytics
                                             SET deleted_at = CURRENT_TIMESTAMP
                                             WHERE id = %s
                                               AND deleted_at IS NULL
                                             """,
                    (analytics_id,),
                )
                logger.info(
                    f"Soft deleted analytics ID {analytics_id}",
                    extra={"analytics_id": analytics_id},
                )
            except Exception as e:
                logger.error(
                    f"Error soft deleting analytics ID {analytics_id}: {e}",
                    extra={"analytics_id": analytics_id, "error": str(e)},
                )
                raise

    async def increment_emoji_reaction(self, clip_id: int, emoji_type: str) -> None:
        """
        Increment the specified emoji reaction count for a clip by 1.
        Creates analytics record if it doesn't exist.

        Args:
            clip_id: The ID of the clip to increment emoji reaction for
            emoji_type: The type of emoji reaction (heart, smile, laugh, sign, surprised)
        """
        with span(
            logger,
            "increment_emoji_reaction",
            {"clip_id": clip_id, "emoji_type": emoji_type},
        ):
            # Map emoji types to column names
            emoji_column_map = {
                "heart": "heart_reactions",
                "smile": "smile_reactions",
                "laugh": "laugh_reactions",
                "sign": "sign_reactions",
                "surprised": "surprised_reactions",
            }

            if emoji_type not in emoji_column_map:
                raise ValueError(f"Invalid emoji type: {emoji_type}")

            column_name = emoji_column_map[emoji_type]

            try:
                # Try to update existing analytics record
                result = await self.db.execute_commit(
                    f"""
                    UPDATE clip_analytics
                    SET {column_name} = {column_name} + 1,
                        updated_at = CURRENT_TIMESTAMP
                    WHERE clip_id = %s
                      AND deleted_at IS NULL
                    """,
                    (clip_id,),
                )

                # If no rows were affected, create a new analytics record
                # execute_commit returns affected row count (or None if unavailable)
                if not result:
                    from common.models.clip import ClipAnalyticsCreate

                    new_analytics = ClipAnalyticsCreate(
                        clip_id=clip_id,
                        views=0,
                        heart_reactions=1 if emoji_type == "heart" else 0,
                        smile_reactions=1 if emoji_type == "smile" else 0,
                        laugh_reactions=1 if emoji_type == "laugh" else 0,
                        sign_reactions=1 if emoji_type == "sign" else 0,
                        surprised_reactions=1 if emoji_type == "surprised" else 0,
                    )
                    await self.create_clip_analytics(new_analytics)
                    logger.info(
                        f"Created new analytics record with {emoji_type} reaction for clip ID {clip_id}"
                    )

            except Exception as e:
                logger.error(
                    f"Error incrementing {emoji_type} reaction for clip ID {clip_id}: {e}",
                    extra={
                        "clip_id": clip_id,
                        "emoji_type": emoji_type,
                        "error": str(e),
                    },
                )
                raise

    async def increment_views(self, clip_id: int) -> None:
        """
        Increment the views count for a clip in clip_analytics, creating
        the analytics row if it does not exist yet.

        Args:
            clip_id: The ID of the clip to increment views for
        """
        with span(logger, "increment_analytics_views", {"clip_id": clip_id}):
            try:
                result = await self.db.execute_commit(
                    """
                    UPDATE clip_analytics
                    SET views = views + 1,
                        updated_at = CURRENT_TIMESTAMP
                    WHERE clip_id = %s
                      AND deleted_at IS NULL
                    """,
                    (clip_id,),
                )

                if not result:
                    # Create analytics row with views=1 when missing
                    from common.models.clip import ClipAnalyticsCreate

                    new_analytics = ClipAnalyticsCreate(
                        clip_id=clip_id,
                        views=1,
                        heart_reactions=0,
                        smile_reactions=0,
                        laugh_reactions=0,
                        sign_reactions=0,
                        surprised_reactions=0,
                    )
                    await self.create_clip_analytics(new_analytics)
                    logger.info(
                        f"Created new analytics record with 1 view for clip ID {clip_id}"
                    )
            except Exception as e:
                logger.error(
                    f"Error incrementing analytics views for clip ID {clip_id}: {e}",
                    extra={"clip_id": clip_id, "error": str(e)},
                )
                raise

    def _row_to_clip_analytics(self, row) -> ClipAnalytics:
        """
        Convert a database row to a ClipAnalytics model.
        """
        return ClipAnalytics(
            id=row[0],
            clip_id=row[1],
            views=row[2],
            heart_reactions=row[3],
            smile_reactions=row[4],
            laugh_reactions=row[5],
            sign_reactions=row[6],
            surprised_reactions=row[7],
            created_at=row[8],
            updated_at=row[9],
            deleted_at=row[10],
        )



"""Pydantic models for video templates."""

from datetime import datetime
from enum import Enum
from typing import Any, List, Optional

from pydantic import BaseModel, Field, ConfigDict


class TierLevel(str, Enum):
    """User tier levels for access control."""

    BASIC = "basic"
    PRO = "pro"
    ENTERPRISE = "enterprise"


class ModifiableFieldType(str, Enum):
    """Types of modifiable fields in templates."""

    TEXT = "text"
    VIDEO = "video"
    IMAGE = "image"
    COLOR = "color"
    FONT = "font"
    NUMBER = "number"
    BOOLEAN = "boolean"


class ModifiableField(BaseModel):
    """Definition of a customizable field in a template."""

    name: str = Field(..., description="Creatomate path (e.g., 'Title.text')")
    label: str = Field(..., description="Human-readable label")
    type: ModifiableFieldType = Field(..., description="Field data type")
    default: Optional[Any] = Field(default=None, description="Default value")
    required: bool = Field(default=False, description="Whether field is required")

    # Type-specific constraints
    max_length: Optional[int] = Field(
        default=None, description="Max length for text fields"
    )
    min_value: Optional[float] = Field(
        default=None, description="Min value for number fields"
    )
    max_value: Optional[float] = Field(
        default=None, description="Max value for number fields"
    )
    allowed_values: Optional[List[str]] = Field(
        default=None, description="Enumerated options"
    )


class VideoTemplateBase(BaseModel):
    """Base fields for video template."""

    creatomate_template_id: str = Field(..., description="Creatomate template ID")
    name: str = Field(..., max_length=255, description="Template name")
    description: Optional[str] = Field(default=None, description="Template description")
    thumbnail_url: Optional[str] = Field(
        default=None, description="Preview thumbnail URL"
    )

    is_global: bool = Field(default=False, description="System-wide template flag")
    tier_required: TierLevel = Field(
        default=TierLevel.BASIC, description="Minimum tier required"
    )
    category: Optional[str] = Field(
        default=None, max_length=100, description="Template category"
    )

    output_width: int = Field(default=1080, gt=0, description="Output width in pixels")
    output_height: int = Field(
        default=1920, gt=0, description="Output height in pixels"
    )
    output_format: str = Field(
        default="mp4", description="Output format (mp4, gif, jpg, png)"
    )

    modifiable_fields: List[ModifiableField] = Field(
        default_factory=list, description="List of customizable fields"
    )


class VideoTemplateCreate(VideoTemplateBase):
    """Data required to create a new video template."""

    user_id: Optional[int] = Field(
        default=None, description="Owner user ID (null for global)"
    )


class VideoTemplateUpdate(BaseModel):
    """Data for updating a video template."""

    name: Optional[str] = Field(default=None, max_length=255)
    description: Optional[str] = None
    thumbnail_url: Optional[str] = None
    tier_required: Optional[TierLevel] = None
    category: Optional[str] = Field(default=None, max_length=100)
    output_width: Optional[int] = Field(default=None, gt=0)
    output_height: Optional[int] = Field(default=None, gt=0)
    output_format: Optional[str] = None
    modifiable_fields: Optional[List[ModifiableField]] = None


class VideoTemplate(VideoTemplateBase):
    """Complete video template model from database."""

    id: int
    user_id: Optional[int] = None
    use_count: int = 0
    created_at: datetime
    updated_at: datetime
    deleted_at: Optional[datetime] = None

    model_config = ConfigDict(from_attributes=True)


class UserTemplatePreferencesBase(BaseModel):
    """Base fields for user template preferences."""

    default_clip_template_id: Optional[int] = Field(
        default=None, description="Default for clips"
    )
    default_highlight_template_id: Optional[int] = Field(
        default=None, description="Default for highlights"
    )
    default_story_template_id: Optional[int] = Field(
        default=None, description="Default for stories"
    )


class UserTemplatePreferencesCreate(UserTemplatePreferencesBase):
    """Data for creating user template preferences."""

    user_id: int


class UserTemplatePreferencesUpdate(UserTemplatePreferencesBase):
    """Data for updating user template preferences."""

    pass


class UserTemplatePreferences(UserTemplatePreferencesBase):
    """Complete user template preferences from database."""

    id: int
    user_id: int
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(from_attributes=True)



from typing import Optional, List
from common.database import DBConfig, Db
from common.logging import get_logger, span

from common.models.tenant_seat import TenantSeat
from common.config.config import ProModelSettings

logger = get_logger(__name__)


class TenantSeatError(Exception):
    """Base exception for tenant seat operations."""


class TenantSeatValidationError(TenantSeatError):
    """Invalid payload or precondition for tenant seat operations."""


class TenantSeatNotFoundError(TenantSeatError):
    """Requested tenant seat subject was not found."""


class TenantSeatForbiddenError(TenantSeatError):
    """Tenant seat action is not permitted."""


class TenantSeatConflictError(TenantSeatError):
    """Tenant seat action conflicts with existing state."""


SELECT_BASE = """
    SELECT id, user_id, email, name, created_at, deleted_at
    FROM tenant_seats
"""

GET_USER_FULL_NAME_QUERY = """
    SELECT first_name, last_name
    FROM users
    WHERE id = %s
    LIMIT 1
"""

GET_USER_TWITCH_USERNAME_QUERY = """
    SELECT twitch_username
    FROM twitch_account
    WHERE user_id = %s
    LIMIT 1
"""

GET_USER_KICK_USERNAME_QUERY = """
    SELECT kick_username
    FROM kick_account
    WHERE user_id = %s
    LIMIT 1
"""

GET_USER_EMAIL_QUERY = """
    SELECT email
    FROM users
    WHERE id = %s
    LIMIT 1
"""

COUNT_ACTIVE_BY_USER_QUERY = """
    SELECT COUNT(*)
    FROM tenant_seats
    WHERE user_id = %s
      AND deleted_at IS NULL
"""

GET_ACTIVE_SEAT_NAME_ONLY_BY_EMAIL_QUERY = """
    SELECT name
    FROM tenant_seats
    WHERE email = %s
      AND deleted_at IS NULL
      AND name IS NOT NULL
      AND btrim(name) <> ''
    ORDER BY created_at DESC, id DESC
    LIMIT 1
"""

SELECT_ACTIVE_BY_USER_QUERY = (
    SELECT_BASE
    + """
    WHERE user_id = %s
      AND deleted_at IS NULL
    ORDER BY created_at DESC, id DESC
"""
)

SOFT_DELETE_BY_ID_QUERY = """
    UPDATE tenant_seats
    SET deleted_at = NOW()
    WHERE id = %s
      AND deleted_at IS NULL
"""

GET_BY_USER_AND_EMAIL_QUERY = (
    SELECT_BASE
    + """
    WHERE user_id = %s
      AND email = %s
    ORDER BY id DESC
    LIMIT 1
"""
)

UPSERT_REVIVE_QUERY = """
    INSERT INTO tenant_seats (user_id, email, name)
    VALUES (%s, %s, %s)
    ON CONFLICT (user_id, email)
    DO UPDATE SET
        name = EXCLUDED.name,
        deleted_at = NULL
    RETURNING id, user_id, email, name, created_at, deleted_at
"""

SELECT_BY_ID_QUERY = (
    SELECT_BASE
    + """
    WHERE id = %s
    LIMIT 1
"""
)


TENANT_SEAT_EXISTS_BY_EMAIL_QUERY = """
    SELECT 1
    FROM tenant_seats
    WHERE email = %s
      AND deleted_at IS NULL
    LIMIT 1
"""


class TenantSeatsDAO:
    """DAO for tenant_seats."""

    def __init__(self, db: Optional[Db] = None):
        self.db = db if db else Db(DBConfig(pg_dsn=None))
        self.proSettings = ProModelSettings()
        logger.debug(
            "TenantSeatsDAO initialized", extra={"component": "TenantSeatsDAO"}
        )

    async def count_by_user(self, user_id: int) -> int:
        """Return how many ACTIVE seats a user has (deleted_at IS NULL)."""
        with span(logger, "ts_count_by_user", {"user_id": user_id}):
            row = await self.db.fetch_one(COUNT_ACTIVE_BY_USER_QUERY, (user_id,))
            if row is None:
                raise RuntimeError(
                    f"Count query returned no row for tenant seats user_id={user_id}"
                )
            try:
                return int(row[0])
            except (TypeError, KeyError):
                raise

    async def get_by_user(self, user_id: int) -> List[TenantSeat]:
        """Return all ACTIVE seats for a user."""
        with span(logger, "ts_get_by_user", {"user_id": user_id}):
            rows = await self.db.fetch_all(SELECT_ACTIVE_BY_USER_QUERY, (user_id,))
            if rows is None:
                return []
            return [self._row_to_model(r) for r in rows]

    async def delete(self, seat_id: int) -> None:
        """Soft delete a seat by id."""
        with span(logger, "ts_delete", {"id": seat_id}):
            await self.db.execute_commit(SOFT_DELETE_BY_ID_QUERY, (seat_id,))
            logger.info("Soft-deleted tenant seat", extra={"id": seat_id})

    async def create(self, user_id: int, email: str, name: str) -> TenantSeat:
        """
        Create or revive seat using UNIQUE(user_id,email).
        - If exists active: updates name (doesn't consume a new seat)
        - If exists deleted or doesn't exist: enforces max_seats, then revives/inserts
        """
        with span(logger, "ts_create", {"user_id": user_id}):
            e = (email or "").strip().lower()
            n = (name or "").strip()

            if not e:
                raise TenantSeatValidationError("email cannot be empty")

            existing = await self.db.fetch_one(
                GET_BY_USER_AND_EMAIL_QUERY, (user_id, e)
            )

            # If seat doesn't exist OR exists but is soft-deleted, it will become ACTIVE => enforce limit
            if not existing or existing[5] is not None:  # deleted_at is at index 5
                current_active = await self.count_by_user(user_id)
                if current_active >= self.proSettings.max_seats:
                    raise TenantSeatConflictError(
                        f"User {user_id} exceeded max_seats ({self.proSettings.max_seats})"
                    )

            # Upsert: pass deleted_at=None so conflict update revives the seat
            columns = ["user_id", "email", "name", "deleted_at"]
            values = (user_id, e, n, None)

            upserted_id = await self.db.upsert(
                table="tenant_seats",
                columns=columns,
                values=values,
                conflict_columns=["user_id", "email"],
                update_columns=["name", "deleted_at"],  # revive + update name
            )

            if not upserted_id:
                raise RuntimeError("Upsert returned no id for tenant_seats")

            row = await self.db.fetch_one(SELECT_BY_ID_QUERY, (upserted_id,))
            if not row:
                raise RuntimeError(f"Could not fetch tenant_seats row id={upserted_id}")

            model = self._row_to_model(row)
            logger.info(
                "Created/revived tenant seat",
                extra={"id": model.id, "user_id": user_id, "email": model.email},
            )
            return model

    def _row_to_model(self, row) -> TenantSeat:
        try:
            return TenantSeat(
                id=row[0],
                user_id=row[1],
                email=row[2],
                name=row[3],
                created_at=row[4],
                deleted_at=row[5],
            )
        except (TypeError, KeyError):
            raise

    async def get_user_full_name(self, user_id: int) -> str:
        """Return best display name for a user: Twitch username, else Kick username, else 'first last'."""
        with span(logger, "ts_get_user_full_name", {"user_id": user_id}):
            row = await self.db.fetch_one(GET_USER_TWITCH_USERNAME_QUERY, (user_id,))
            twitch_username = (row[0] or "").strip() if row else ""
            if twitch_username:
                return twitch_username

            row = await self.db.fetch_one(GET_USER_KICK_USERNAME_QUERY, (user_id,))
            kick_username = (row[0] or "").strip() if row else ""
            if kick_username:
                return kick_username

            row = await self.db.fetch_one(GET_USER_FULL_NAME_QUERY, (user_id,))
            if not row:
                raise TenantSeatNotFoundError(f"User {user_id} not found")

            first = (row[0] or "").strip()
            last = (row[1] or "").strip()
            full = f"{first} {last}".strip()

            if not full:
                raise TenantSeatValidationError(f"User {user_id} has no name on record")

            return full

    async def update_seat_name(self, user_id: int, email: str, name: str) -> TenantSeat:
        """Update seat name for (user_id, email)."""
        with span(logger, "ts_update_seat_name", {"user_id": user_id}):
            e = (email or "").strip().lower()
            n = (name or "").strip()

            if not e:
                raise TenantSeatValidationError("email cannot be empty")
            if not n:
                raise TenantSeatValidationError("name cannot be empty")

            row = await self.db.fetch_one(GET_BY_USER_AND_EMAIL_QUERY, (user_id, e))
            if not row:
                raise TenantSeatNotFoundError("Seat not found")

            # Only allow updates for active seats
            if row[5] is not None:  # deleted_at index
                raise TenantSeatNotFoundError("Seat is deleted")

            seat_id = int(row[0])

            await self.db.update(
                table="tenant_seats",
                set_columns=["name"],
                set_values=[n],
                where_clause="id = %s AND deleted_at IS NULL",
                where_params=(seat_id,),
            )

            updated = await self.db.fetch_one(SELECT_BY_ID_QUERY, (seat_id,))
            if not updated:
                raise RuntimeError(f"Could not fetch tenant_seats row id={seat_id}")

            return self._row_to_model(updated)

    async def get_user_email(self, user_id: int) -> str:
        """Return the user's email from users table."""
        with span(logger, "ts_get_user_email", {"user_id": user_id}):
            row = await self.db.fetch_one(GET_USER_EMAIL_QUERY, (user_id,))
            if not row or not row[0]:
                raise TenantSeatNotFoundError(f"User {user_id} not found")
            return str(row[0]).strip().lower()

    async def get_seat_name_by_email(self, email: str) -> str:
        """Return tenant_seats.name for an active seat by email (non-empty)."""
        with span(logger, "ts_get_seat_name_by_email", {"email": email}):
            e = (email or "").strip().lower()
            if not e:
                raise TenantSeatValidationError("email cannot be empty")

            row = await self.db.fetch_one(
                GET_ACTIVE_SEAT_NAME_ONLY_BY_EMAIL_QUERY, (e,)
            )
            name = (row[0] or "").strip() if row else ""
            if not name:
                return ""

            return name

    async def seat_exists_by_email(self, email: str) -> bool:
        """Return True if an ACTIVE tenant seat exists for the given email."""
        with span(logger, "ts_seat_exists_by_email", {"email": email}):
            e = (email or "").strip().lower()
            if not e:
                raise TenantSeatValidationError("email cannot be empty")

            row = await self.db.fetch_one(TENANT_SEAT_EXISTS_BY_EMAIL_QUERY, (e,))
            return row is not None



"""Common exceptions for the TL;DR.tv platform."""


class CommonError(Exception):
    """Base exception for all common module errors."""

    pass


class DatabaseError(CommonError):
    """Raised when database operations fail."""

    pass


class PlatformError(CommonError):
    """Base exception for platform-specific errors."""

    pass


class TwitchError(PlatformError):
    """Raised when Twitch API operations fail."""

    pass


class KickError(PlatformError):
    """Raised when Kick API operations fail."""

    pass


class ValidationError(CommonError):
    """Raised when data validation fails."""

    pass


class AuthenticationError(CommonError):
    """Raised when authentication fails."""

    pass


class RateLimitError(PlatformError):
    """Raised when API rate limits are exceeded."""

    pass


class WebhookError(PlatformError):
    """Raised when webhook operations fail."""

    pass


class MessageQueueError(CommonError):
    """Raised when message queue operations fail."""

    pass



import datetime
from typing import List, Optional

from pydantic import BaseModel, EmailStr, ConfigDict, Field
from common.models.common import Platform, UserAccessToken


class KickAccountBase(BaseModel):
    """Model for Kick account data."""

    user_id: int
    kick_id: str
    kick_username: str
    profile_link: Optional[str] = None
    profile_picture: Optional[str] = None
    authorized_at: Optional[datetime.datetime] = None


class KickAccountCreate(KickAccountBase):
    """Model for creating a new Kick account."""

    pass


class KickAccountUpdate(BaseModel):
    """Model for updating Kick account data."""

    user_id: Optional[int] = None
    kick_id: Optional[str] = None
    kick_username: Optional[str] = None
    profile_link: Optional[str] = None
    profile_picture: Optional[str] = None
    updated_at: Optional[datetime.datetime] = Field(
        default_factory=lambda: datetime.datetime.now(datetime.UTC)
    )
    deleted_at: Optional[datetime.datetime] = None
    authorized_at: Optional[datetime.datetime] = None


class KickAccount(KickAccountBase):
    """Complete Kick account model including database ID."""

    id: int
    access_token: Optional[UserAccessToken] = None
    created_at: datetime.datetime
    updated_at: datetime.datetime
    deleted_at: Optional[datetime.datetime] = None
    authorized_at: Optional[datetime.datetime] = None

    model_config = ConfigDict(from_attributes=True)


class TwitchAccountBase(BaseModel):
    """Model for Twitch account data."""

    user_id: int
    twitch_id: str
    twitch_username: str
    profile_link: Optional[str] = None
    profile_picture: Optional[str] = None
    authorized_at: Optional[datetime.datetime] = None
    oauth_token_id: Optional[int] = None


class SocialMediaAccountCreate(BaseModel):
    """Model for Social Media account data creation."""

    platform: Platform
    user_id: int
    oauth_token_id: Optional[int] = None
    username: Optional[str] = None
    authorized_at: Optional[datetime.datetime] = None
    created_at: Optional[datetime.datetime] = None
    updated_at: Optional[datetime.datetime] = None


class SocialMediaAccount(SocialMediaAccountCreate):
    """Model for Social Media account data."""

    id: int


class TwitchAccountCreate(TwitchAccountBase):
    """Model for creating a new Twitch account."""

    pass


class TwitchAccountUpdate(BaseModel):
    """Model for updating Twitch account data."""

    user_id: Optional[int] = None
    twitch_id: Optional[str] = None
    twitch_username: Optional[str] = None
    profile_link: Optional[str] = None
    profile_picture: Optional[str] = None
    updated_at: Optional[datetime.datetime] = Field(
        default_factory=lambda: datetime.datetime.now(datetime.UTC)
    )
    deleted_at: Optional[datetime.datetime] = None
    authorized_at: Optional[datetime.datetime] = None
    oauth_token_id: Optional[int] = None


class TwitchAccount(TwitchAccountBase):
    id: int
    access_token: Optional[UserAccessToken] = None
    created_at: datetime.datetime
    updated_at: datetime.datetime
    deleted_at: Optional[datetime.datetime] = None

    model_config = ConfigDict(from_attributes=True)


class UserBase(BaseModel):
    """Base model for user data."""

    email: EmailStr
    first_promoter_id: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    phone: Optional[str] = None
    tier: str = "basic"
    product: Optional[str] = None


class UserCreate(UserBase):
    """Model for creating a new user."""

    pass


class UserUpdate(BaseModel):
    """Model for updating user data."""

    email: Optional[EmailStr] = None
    first_promoter_id: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    phone: Optional[str] = None
    tier: Optional[str] = None
    product: Optional[str] = None
    updated_at: Optional[datetime.datetime] = Field(
        default_factory=lambda: datetime.datetime.now(datetime.UTC)
    )
    last_login: Optional[datetime.datetime] = None
    customer_journey_completed_at: Optional[datetime.datetime] = None
    deleted_at: Optional[datetime.datetime] = None


class User(UserBase):
    """Complete user model including database ID."""

    id: int
    twitch_account: Optional[TwitchAccount] = None
    kick_account: Optional[KickAccount] = None
    created_at: datetime.datetime
    updated_at: datetime.datetime
    deleted_at: Optional[datetime.datetime] = None
    last_login: Optional[datetime.datetime] = None
    customer_journey_completed_at: Optional[datetime.datetime] = None
    social_media_platforms: Optional[List[str]] = None
    is_tenant_user: Optional[bool] = None

    model_config = ConfigDict(from_attributes=True)



from typing import Optional

from dotenv import find_dotenv
from pydantic_settings import BaseSettings, SettingsConfigDict


class LogFireSettings(BaseSettings):
    log_env: str = "development"
    logfire_write_token: Optional[str] = None

    model_config = SettingsConfigDict(
        env_file=find_dotenv(), env_file_encoding="utf-8", extra="ignore"
    )



# common/pulsar_client.py

from dotenv import find_dotenv
from pydantic_settings import BaseSettings, SettingsConfigDict


class MQConfig(BaseSettings):
    """
    Configuration settings for the Message Queue.
    """

    mq_service_url: str = "amqp://guest:guest@localhost:5672"
    mq_username: str = "guest"
    mq_password: str = "guest"

    model_config = SettingsConfigDict(
        env_file=find_dotenv(), env_file_encoding="utf-8", extra="ignore"
    )



from abc import ABC, abstractmethod
from typing import Any, List, Sequence

from common.models.common import ChatMessage


class ChatMessageDao(ABC):
    """
    Storage-agnostic contract for chat-message persistence/querying.
    Concrete implementations must fulfil the methods declared here.
    """

    # ------------------------------------------------------------------ #
    # CRUD-like operations                                               #
    # ------------------------------------------------------------------ #

    @abstractmethod
    async def save_message(self, message: ChatMessage) -> str | None:
        """Persist a single chat message."""
        raise NotImplementedError

    @abstractmethod
    async def delete_message(self, message_id: int | str) -> bool | None:
        """Remove an already-stored message."""
        raise NotImplementedError

    @abstractmethod
    async def get_message_by_id(self, message_id: int | str) -> ChatMessage | None:
        """Fetch a message by its primary identifier."""
        raise NotImplementedError

    # ------------------------------------------------------------------ #
    # Query helpers                                                      #
    # ------------------------------------------------------------------ #

    @abstractmethod
    async def get_messages_by_time_range(
        self, broadcaster_user_id: str, start_timestamp: float, end_timestamp: float
    ) -> Sequence[ChatMessage]:
        """All messages whose timestamp is between `start` and `end`."""
        raise NotImplementedError

    @abstractmethod
    async def get_messages_by_broadcaster(
        self,
        broadcaster_id: int | str,
        limit: int | None = None,
    ) -> Sequence[ChatMessage]:
        """Messages that belong to the given broadcaster."""
        raise NotImplementedError

    @abstractmethod
    async def get_messages_by_chatter(
        self,
        chatter_id: int | str,
        limit: int | None = None,
    ) -> Sequence[ChatMessage]:
        """Messages posted by a particular user/chatter."""
        raise NotImplementedError

    @abstractmethod
    async def get_messages_by_stream(
        self,
        stream_id: int | str,
        limit: int | None = None,
    ) -> Sequence[ChatMessage]:
        """Messages associated with a specific stream."""
        raise NotImplementedError

    @abstractmethod
    async def count_messages_by_stream(self, stream_id: int | str) -> int:
        """Number of messages posted during a particular stream."""
        raise NotImplementedError

    # ------------------------------------------------------------------ #
    # Internal / factory helpers                                         #
    # ------------------------------------------------------------------ #

    @abstractmethod
    def _build_messages_from_rows(
        self,
        rows: List[tuple[Any, ...]],
    ) -> List[ChatMessage]:
        """Convert raw DB rows into domain objects."""
        raise NotImplementedError

    @abstractmethod
    def _build_message_from_data(
        self,
        data: dict[str, Any] | tuple[Any, ...],
    ) -> ChatMessage:
        """Create a domain object from a raw mapping/dict."""
        raise NotImplementedError

    @abstractmethod
    def _to_jsonb(self, payload: Any) -> Any:
        """
        Convert a Python mapping to a driver-specific JSONB representation.
        For non-SQL stores this can simply return the original dict.
        """
        raise NotImplementedError



from .model import (
    TldrMessage,
    UserSignupMessage,
    StreamOnlineMessage,
    StreamOfflineMessage,
    VoiceClipMessage,
)
from .mq_config import MQConfig

__all__ = [
    "MQConfig",
    "TldrMessage",
    "UserSignupMessage",
    "StreamOnlineMessage",
    "StreamOfflineMessage",
    "VoiceClipMessage",
]



# models/twitch_models.py
from datetime import datetime
from typing import List, Optional, Dict, Any

from pydantic import BaseModel, Field

import common.models.common


class MarkStreamRequest(BaseModel):
    user_id: str
    description: Optional[str] = None


class MarkStreamResponse(BaseModel):
    id: str
    created_at: datetime
    position_seconds: int
    description: str


class StreamMarker(BaseModel):
    id: str
    created_at: str
    description: str
    position_seconds: int
    URL: Optional[str] = None


class VideoMarkers(BaseModel):
    video_id: str
    markers: List[StreamMarker]


class UserMarkers(BaseModel):
    user_id: str
    user_name: str
    user_login: str
    videos: List[VideoMarkers]


class GetStreamMarkersResponse(BaseModel):
    data: List[UserMarkers]
    pagination: dict = Field(default_factory=dict)


class User(BaseModel):
    id: str
    login: str
    display_name: str
    type: str
    broadcaster_type: str
    description: str
    profile_image_url: str
    offline_image_url: str
    view_count: int
    email: Optional[str] = None  # Available if scopes include user:read:email


class ChannelInfo(BaseModel):
    broadcaster_id: str
    broadcaster_login: str
    broadcaster_name: str
    broadcaster_language: str
    game_id: str
    game_name: str
    title: str
    delay: int
    tags: List[str]
    content_classification_labels: List[str]
    is_branded_content: bool


class Stream(BaseModel):
    id: str
    user_id: str
    user_login: str
    user_name: str
    game_id: str
    game_name: str
    type: str
    title: str
    viewer_count: int
    started_at: str
    language: str
    thumbnail_url: str
    tag_ids: List[str]
    is_mature: bool


class EventSubTransport(BaseModel):
    method: str
    callback: Optional[str] = None
    secret: Optional[str] = None
    session_id: Optional[str] = None  # For WebSocket transports
    conduit_id: Optional[str] = None  # For Conduit transports


class EventSubSubscriptionRequest(BaseModel):
    type: str
    version: str
    condition: Dict[str, Any]
    transport: EventSubTransport


class EventSubSubscription(BaseModel):
    id: str
    status: str
    type: str
    version: str
    condition: Dict[str, Any]
    transport: Dict[str, Any]
    created_at: str
    cost: int


class EventSubNotification(BaseModel):
    subscription: EventSubSubscription
    event: Dict[str, Any]  # Structure depends on the subscription type


# New Models for Conduits
class EventSubConduit(BaseModel):
    id: str
    shard_count: int


class EventSubConduitShard(BaseModel):
    id: str
    transport: EventSubTransport
    status: Optional[str] = None
    connected_at: Optional[str] = None
    disconnected_at: Optional[str] = None


class EventSubConduitShardsUpdate(BaseModel):
    conduit_id: str
    shards: List[EventSubConduitShard]


class EventSubConduitShardsUpdateResponse(BaseModel):
    data: List[EventSubConduitShard]


class WebSocketSession(BaseModel):
    id: str
    status: str
    connected_at: str
    keepalive_timeout_seconds: Optional[int] = None
    reconnect_url: Optional[str] = None


class WebSocketMessageMetadata(BaseModel):
    message_id: str
    message_type: str
    message_timestamp: str
    subscription_type: Optional[str] = None
    subscription_version: Optional[str] = None


class WebSocketMessagePayload(BaseModel):
    subscription: Optional[EventSubSubscription] = None
    event: Optional[Dict[str, Any]] = None
    session: Optional[WebSocketSession] = None


class WebSocketMessage(BaseModel):
    metadata: WebSocketMessageMetadata
    payload: WebSocketMessagePayload


class TwitchChatMessageRequest(BaseModel):
    broadcaster_id: str
    sender_id: str
    message: str = Field(max_length=500)
    reply_parent_message_id: Optional[str] = None


class TwitchChatMessageDropReason(BaseModel):
    code: str
    message: str


class TwitchChatMessageResponse(BaseModel):
    message_id: str
    is_sent: bool
    drop_reason: Optional[TwitchChatMessageDropReason] = None


class MessageFragment(BaseModel):
    type: str
    text: str
    cheermote: Optional[Dict[str, Any]] = None
    emote: Optional[Dict[str, Any]] = None
    mention: Optional[Dict[str, Any]] = None


class Badge(BaseModel):
    set_id: str
    id: str
    info: str = ""


class MessageContent(BaseModel):
    text: str
    fragments: List[MessageFragment]


class TwitchChatMessage(common.models.common.ChatMessage):
    # Broadcaster info
    broadcaster_user_id: str
    broadcaster_user_login: str
    broadcaster_user_name: str

    # Chatter info
    chatter_user_id: str
    chatter_user_login: str
    chatter_user_name: str

    # Message data
    message_id: str
    message: MessageContent

    # Visual styles
    color: Optional[str] = None
    badges: List[Badge] = []

    # Message metadata
    message_type: str = "text"
    created_at: float

    # Additional fields
    cheer: Optional[Dict[str, Any]] = None
    reply: Optional[Dict[str, Any]] = None
    channel_points_custom_reward_id: Optional[str] = None

    # Source info (for replies/shares)
    source_broadcaster_user_id: Optional[str] = None
    source_broadcaster_user_login: Optional[str] = None
    source_broadcaster_user_name: Optional[str] = None
    source_message_id: Optional[str] = None
    source_badges: Optional[List[Badge]] = None

    def get_content(self) -> str:
        """Get the content of the chat message."""
        return self.message.text

    def get_timestamp(self) -> float:
        """Get the timestamp of the chat message."""
        return self.created_at

    def get_broadcaster_user_id(self) -> str:
        """Get the broadcaster user ID."""
        return self.broadcaster_user_id

    @classmethod
    def from_eventsub(cls, event: Dict[str, Any]) -> "TwitchChatMessage":
        """Create a TwitchChatMessage from an EventSub message event"""
        message_data = event.get("message", {})

        # Convert raw fragments to MessageFragment objects
        fragments = [
            MessageFragment(
                type=frag.get("type", "text"),
                text=frag.get("text", ""),
                cheermote=frag.get("cheermote"),
                emote=frag.get("emote"),
                mention=frag.get("mention"),
            )
            for frag in message_data.get("fragments", [])
        ]

        # Create MessageContent
        message = MessageContent(text=message_data.get("text", ""), fragments=fragments)

        # Convert badges
        badges = [
            Badge(
                set_id=badge.get("set_id", ""),
                id=badge.get("id", ""),
                info=badge.get("info", ""),
            )
            for badge in event.get("badges", [])
        ]

        # Convert source badges if present
        source_badges = None
        if event.get("source_badges"):
            source_badges = [
                Badge(
                    set_id=badge.get("set_id", ""),
                    id=badge.get("id", ""),
                    info=badge.get("info", ""),
                )
                for badge in event.get("source_badges", [])
            ]

        # cls.Meta.database = connect()

        c = cls(
            broadcaster_user_id=event.get("broadcaster_user_id"),
            broadcaster_user_login=event.get("broadcaster_user_login"),
            broadcaster_user_name=event.get("broadcaster_user_name"),
            chatter_user_id=event.get("chatter_user_id"),
            chatter_user_login=event.get("chatter_user_login"),
            chatter_user_name=event.get("chatter_user_name"),
            message_id=event.get("message_id"),
            message=message,
            color=event.get("color"),
            badges=badges,
            message_type=event.get("message_type", "text"),
            cheer=event.get("cheer"),
            reply=event.get("reply"),
            channel_points_custom_reward_id=event.get(
                "channel_points_custom_reward_id"
            ),
            source_broadcaster_user_id=event.get("source_broadcaster_user_id"),
            source_broadcaster_user_login=event.get("source_broadcaster_user_login"),
            source_broadcaster_user_name=event.get("source_broadcaster_user_name"),
            source_message_id=event.get("source_message_id"),
            source_badges=source_badges,
            created_at=datetime.now().timestamp(),
        )
        return c



from .clip import ClipDAO as ClipDAO, ClipAnalyticsDAO as ClipAnalyticsDAO
from .custom_wake_words import CustomWakeWordsDAO as CustomWakeWordsDAO
from .stream import StreamDAO as StreamDAO
from .stream_key import StreamKeyDAO as StreamKeyDAO
from .twitch_chat_message import (
    TwitchChatMessageDAO as TwitchChatMessageDAO,
)
from .upload_session import UploadSessionDAO as UploadSessionDAO
from .user import UserDAO as UserDAO
from .scheduled_post import ScheduledPostDAO as ScheduledPostDAO
from .timestamp import StreamDeckSignalDAO
from .video_template import VideoTemplateDAO as VideoTemplateDAO

__all__ = [
    "ClipDAO",
    "ClipAnalyticsDAO",
    "CustomWakeWordsDAO",
    "StreamDAO",
    "StreamKeyDAO",
    "TwitchChatMessageDAO",
    "UploadSessionDAO",
    "UserDAO",
    "ScheduledPostDAO",
    "StreamDeckSignalDAO",
    "VideoTemplateDAO",
]



import colorlog


def get_formatter():
    return colorlog.ColoredFormatter(
        "%(asctime)s %(log_color)s%(bold)s%(levelname)-5s%(reset)s [%(process)d] %(name)s:%(lineno)d | %(message)s",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
    )



from typing import Dict, List, Optional

import httpx

from common import logging

logger = logging.get_logger(__name__)


class FacebookPublisher:
    FB_API_VERSION = "v23.0"
    GRAPH = f"https://graph.facebook.com/{FB_API_VERSION}"
    GRAPH_VIDEO = f"https://graph-video.facebook.com/{FB_API_VERSION}"

    @staticmethod
    async def get_user_id(access_token: str) -> str:
        url = f"{FacebookPublisher.GRAPH}/me"
        params = {"fields": "id,name", "access_token": access_token}
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(url, params=params)
                response.raise_for_status()
                data = response.json()
                user_id = data.get("id")
                if not user_id:
                    raise RuntimeError("Error fetching user id from Facebook API.")
                return user_id
            except httpx.HTTPStatusError as e:
                logger.error(
                    f"[Facebook] Error HTTP {e.response.status_code}: {e.response.text}"
                )
                raise
            except Exception as e:
                logger.error(f"[Facebook] Error fetching user_id: {e}")
                raise

    @staticmethod
    async def get_user_pages_tokens(
        user_id: str, access_token: str
    ) -> List[Dict[str, str]]:
        url = f"{FacebookPublisher.GRAPH}/{user_id}/accounts"
        params = {"fields": "id,access_token", "access_token": access_token}
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(url, params=params)
                response.raise_for_status()
                data = response.json()
                if "data" not in data:
                    raise RuntimeError(
                        "Unexpected Facebook API response: missing 'data' field."
                    )
                pages_info = []
                for page in data["data"]:
                    page_id = page.get("id")
                    page_token = page.get("access_token")
                    if not (page_id and page_token):
                        continue
                    pages_info.append({"page_id": page_id, "page_token": page_token})
                return pages_info
            except httpx.HTTPStatusError as e:
                logger.error(
                    f"[Facebook API] HTTP {e.response.status_code}: {e.response.text}"
                )
                raise
            except RuntimeError as e:
                logger.error(f"[Facebook API] Data error: {e}")
                raise
            except Exception as e:
                logger.error(
                    f"[Facebook API] Unexpected error while fetching page tokens: {e}"
                )
                raise

    @staticmethod
    async def publish_video(
        page_id: str,
        page_access_token: str,
        video_url: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> str:
        url = f"{FacebookPublisher.GRAPH_VIDEO}/{page_id}/videos"
        data = {"access_token": page_access_token, "file_url": video_url}
        if title:
            data["title"] = title
        if description:
            data["description"] = description
        async with httpx.AsyncClient(timeout=None) as client:
            resp = await client.post(url, data=data)
            resp.raise_for_status()
            payload = resp.json()
            return payload["id"]

    @staticmethod
    async def publish_video_story(
        page_id: str,
        page_access_token: str,
        video_url: str,
    ) -> dict:
        start_url = f"{FacebookPublisher.GRAPH}/{page_id}/video_stories"
        start_params = {
            "upload_phase": "start",
            "access_token": page_access_token,
        }

        async with httpx.AsyncClient(timeout=None) as client:
            start_resp = await client.post(start_url, params=start_params)
            start_resp.raise_for_status()
            start_payload = start_resp.json()

            video_id = start_payload["video_id"]
            upload_url = start_payload["upload_url"]

            upload_headers = {
                "Authorization": f"OAuth {page_access_token}",
                "file_url": video_url,
            }
            upload_resp = await client.post(upload_url, headers=upload_headers)
            upload_resp.raise_for_status()

            finish_params = {
                "upload_phase": "finish",
                "video_id": video_id,
                "access_token": page_access_token,
            }
            finish_resp = await client.post(start_url, params=finish_params)
            finish_resp.raise_for_status()
            return finish_resp.json()
